""" pynchon.plugins.parse
"""

import shimport

from pynchon import abcs, events, models  # noqa
from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)
config = shimport.lazy("pynchon.config")


class Parse(models.ToolPlugin):
    """
    Misc tools for parsing
    """

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "parse"

    name = "parse"
    cli_name = "parse"
    cli_aliases = ["parser"]



""" pynchon.plugins.fixme
"""

from fnmatch import fnmatch

from pynchon import abcs, constants, models
from pynchon.util import lme, typing
from pynchon.util.os import invoke

from fleks.cli import click, options  # noqa


LOGGER = lme.get_logger(__name__)


class FixMeConfig(abcs.Config):
    config_key: typing.ClassVar[str] = "fixme"
    exclude_patterns: typing.List[str] = typing.Field()

    # @tagging.tagged_property(conflict_strategy="override")
    @property
    def exclude_patterns(self):
        from pynchon.config import globals

        global_ex = globals.exclude_patterns
        my_ex = self.__dict__.get("exclude_patterns", [])
        return list(set(global_ex + my_ex))


class FixMe(models.Planner):
    """Generates {docs_root}/FIXME.md from source"""

    name = "fixme"
    config_class = FixMeConfig

    def plan(self, config: dict = None) -> typing.List:
        """ """
        config = config or self.__class__.get_current_config()
        plan = super(self.__class__, self).plan(config)
        target = abcs.Path(self.project_config["docs"]["root"]) / "FIXME.md"
        plan.append(
            self.goal(
                type="gen",
                resource=target,
                command=f"pynchon fixme gen --output {target}",
            )
        )
        return plan

    @click.option(
        "--output",
        "-o",
        default="docs/FIXME.md",
        help=("output file to write.  (optional)"),
    )
    @options.should_print
    @options.header
    def gen(
        self,
        output,
        should_print,
        header,
    ):
        """Generate FIXME.md files, aggregating references to all FIXME's in code-base

        :param output: param should_print:
        :param header:
        :param should_print:

        """
        from pynchon import api

        config = self.__class__.project_config
        src_root = config.src["root"]
        exclude_patterns = self["exclude_patterns"]
        cmd = invoke(f"grep --line-number -R FIXME {src_root}")
        assert cmd.succeeded
        items = []
        skipped = {}
        for line in cmd.stdout.split("\n"):
            line = line.strip()
            if not line or line.startswith("Binary file"):
                continue
            bits = line.split(":")
            file = bits.pop(0)
            path = abcs.Path(file)
            for g in exclude_patterns:
                if fnmatch(file, g):
                    skipped[g] = skipped.get(g, []) + [file]
                    break
            else:
                line_no = bits.pop(0)
                items.append(
                    dict(
                        file=abcs.Path(file).relative_to(
                            abcs.Path(config["git"]["root"]).absolute()
                        ),
                        line=":".join(bits),
                        line_no=line_no,
                    )
                )
        for g in skipped:
            msg = f"exclude_pattern @ `{g}` skipped {len(skipped[g])} matches"
            LOGGER.info(msg)
        result = api.render.get_template(
            "pynchon/plugins/fixme/FIXME.md.j2",
        ).render(**dict(items=items))
        msg = result
        print(msg, file=open(output, "w"))
        if should_print and output != "/dev/stdout":
            print(msg)

    @classmethod
    def asdasdinit_cli(kls):
        """

        :param kls:

        """
        parent = kls.click_group
        T_FIXME = constants.ENV.get_template(
            "pynchon/plugins/plugins/fixme/FIXME.md.j2"
        )



""" {{pkg}}.util.lme
"""

import logging

from rich.style import Style
from rich.theme import Theme
from rich.console import Console
from rich.logging import RichHandler
from rich.default_styles import DEFAULT_STYLES

from pynchon import constants

THEME = Theme(
    {
        **DEFAULT_STYLES,
        **{
            "logging.keyword": Style(bold=True, color="yellow"),
            # "logging.level.notset": Style(dim=True),
            "logging.level.debug": Style(color="green"),
            "logging.level.info": Style(
                dim=True,
                # color="blue",
            ),
            "logging.level.warning": Style(color="yellow"),
            "logging.level.error": Style(color="red", dim=True, bold=True),
            "logging.level.critical": Style(
                color="red",
                bold=True,
                # reverse=True
            ),
            "log.level": Style.null(),
            "log.time": Style(color="cyan", dim=True),
            "log.message": Style.null(),
            "log.path": Style(dim=True),
        },
    }
)
CONSOLE = Console(theme=THEME, stderr=True)


def set_global_level(level):
    """https://stackoverflow.com/questions/19617355/dynamically-changing-log-level-without-restarting-the-application

    :param level:

    """
    logger = logging.getLogger()
    logger.setLevel(level)
    for handler in logger.handlers:
        handler.setLevel(level)


class Fake:
    warning = debug = info = critical = lambda *args, **kwargs: None
    # if isinstance(handler, type(logging.StreamHandler())):
    #     handler.setLevel(logging.DEBUG)
    #     logger.debug('Debug logging enabled')


def get_logger(name, console=CONSOLE, fake=False):
    """utility function for returning a logger
    with standard formatting patterns, etc

    :param name: param console:  (Default value = CONSOLE)
    :param console:  (Default value = CONSOLE)

    """
    if fake:
        return Fake()
    log_handler = RichHandler(
        rich_tracebacks=True,
        console=CONSOLE,
        show_time=False,
    )

    logging.basicConfig(
        format="%(message)s",
        datefmt="[%X]",
        handlers=[log_handler],
    )
    FormatterClass = logging.Formatter
    formatter = FormatterClass(
        fmt=" ".join(["%(name)s", "%(message)s"]),
        # datefmt="%Y-%m-%d %H:%M:%S",
        datefmt="",
    )
    log_handler.setFormatter(formatter)

    logger = logging.getLogger(name)

    # FIXME: get this from some kind of global config
    # logger.setLevel("DEBUG")
    logger.setLevel(constants.LOG_LEVEL.upper())

    return logger



""" pynchon.api.render

    Basically this is core jinja stuff.
    Looking for CLI entrypoints?  See pynchon.util.text.render
    Looking for the JinjaPlugin?  See pynchon.plugins.jinja
"""

import os
import functools

from jinja2 import Environment  # Template,; UndefinedError,
from jinja2 import FileSystemLoader, StrictUndefined

import pynchon
from pynchon import abcs, constants, events
from pynchon.util.os import invoke

import jinja2  # noqa

from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


def is_templated(txt: str = "") -> bool:
    """ """
    return txt is not None and ("{{" in txt and "}}" in txt)


def dictionary(input, context):
    """
    :param context:
    """
    from pynchon.abcs.visitor import JinjaDict

    return JinjaDict(input).render(context)


@functools.lru_cache(maxsize=None)
def get_jinja_globals():
    """ """
    events.lifecycle.send(__name__, msg="finalizing jinja globals")

    def invoke_helper(*args, **kwargs) -> typing.StringMaybe:
        """A jinja filter/extension"""
        out = invoke(*args, **kwargs)
        assert out.succeeded
        return out.stdout

    def markdown_toc(fname: str, level=None):
        """ """
        assert fname
        fname = abcs.Path(fname)
        assert fname.exists()
        script = abcs.Path(pynchon.__file__).parents[0] / "scripts" / "gh-md-toc.sh"
        result = invoke(
            f"cat {fname} | bash {script} -", command_logger=LOGGER.critical
        )
        assert result.succeeded
        return result.stdout

    return dict(
        sh=invoke_helper,
        bash=invoke_helper,
        invoke=invoke_helper,
        map=map,
        markdown_toc=markdown_toc,
        eval=eval,
        env=os.getenv,
    )


def get_jinja_includes(*includes):
    """ """
    includes = list(includes)
    includes += list(constants.PYNCHON_CORE_INCLUDES_DIRS)

    return [abcs.Path(t) for t in includes]


@functools.lru_cache(maxsize=None)
def get_jinja_env(*includes, quiet: bool = False):
    """ """
    events.lifecycle.send(__name__, msg="finalizing jinja-Env")
    includes = get_jinja_includes(*includes)
    for template_dir in includes:
        if not template_dir.exists:
            err = f"template directory @ `{template_dir}` does not exist"
            raise ValueError(err)
    # includes and (not quiet) and LOGGER.warning(f"Includes: {includes}")
    env = Environment(
        loader=FileSystemLoader([str(t) for t in includes]),
        undefined=StrictUndefined,
    )
    env.pynchon_includes = includes

    env.globals.update(**get_jinja_globals())

    known_templates = list(map(abcs.Path, set(env.loader.list_templates())))

    if known_templates:
        from pynchon.util import text as util_text

        msg = "Known template search paths (includes folders only): "
        tmp = list({p.parents[0] for p in known_templates})
        LOGGER.info(msg + util_text.to_json(tmp))
    return env


def get_template_from_string(
    content,
    env=None,
):
    """
    :param env=None:
    :param content:
    """
    env = env or get_jinja_env()
    return env.from_string(content)


def get_template_from_file(
    file: str = None,
    **kwargs,
):
    """
    :param file: str = None:
    :param **kwargs:
    """
    with open(file) as fhandle:
        content = fhandle.read()
    return get_template_from_string(content, **kwargs)


def get_template(
    template_name: typing.Union[str, abcs.Path] = None,
    env=None,
    from_string: str = None,
):
    """ """
    env = env or get_jinja_env()
    if isinstance(template_name, (abcs.Path,)):
        template_name = str(template_name)
    try:
        if from_string:
            template = env.from_string(from_string)
        else:
            LOGGER.info(f"Looking up {template_name}")
            template = env.get_template(template_name)
    except (jinja2.exceptions.TemplateNotFound,) as exc:
        LOGGER.critical(f"Template exception: {exc}")
        LOGGER.critical(f"Jinja-includes: {env.pynchon_includes}")
        err = getattr(exc, "templates", exc.message)
        LOGGER.critical(f"Problem template: {err}")
        raise
    template.render = functools.partial(template.render, __template__=template_name)
    return template


def clean_whitespace(txt: str):
    # return txt
    return "\n".join([x for x in txt.split("\n") if x.strip()])



""" pynchon.testing

WARNING: this file exists so that some things can be shared across
the different test suites (units, integration, smoke-tests) which
are defined by this project.  To avoid more WETness in your
`{suite_name}/conftest.py`, lift common values up into this file.
"""



""" pynchon.api.project
"""

from pynchon.util import lme, text, typing

LOGGER = lme.get_logger(__name__)


def get_config() -> dict:
    """ """
    from pynchon.config.util import finalize

    return finalize()


def plan(config: dict = {}) -> dict:
    """
    :param config: dict:  (Default value = {})
    """
    plan = []
    config = config or get_config()
    project = config["project"]
    from pynchon.plugins.util import get_plugin_obj

    for plugin_name in config["pynchon"]["plugins"]:
        plugin = get_plugin_obj(plugin_name)
        if not plugin.contribute_plan_apply:
            continue
        fxn = getattr(plugin, "plan", None)
        if fxn is None:
            continue
        plugin.logger.debug("Planning..")
        result = plugin.plan(config)
        msg = "Done planning.  "
        if result:
            msg = msg + "result={}"
            plugin.logger.debug(msg.format(text.to_json(result)))
        else:
            plugin.logger.critical(msg + "But plugin produced an empty plan!")
        assert isinstance(
            result, typing.List
        ), f"plugin @ {plugin_name} generates bad plan {result}"
        plan += result

    return config, plan



""" pynchon.util.os.pids
"""

import shimport

from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)

psutil = shimport.lazy("psutil")


def filter_pids(cmdline__contains: str = None, **kwargs):
    """ """
    procs = []
    for p in psutil.pids():
        try:
            procs.append(psutil.Process(p))
        except (psutil.NoSuchProcess,) as exc:
            LOGGER.critical(exc)
    survivors = []
    for p in procs:
        try:
            # special filters
            if cmdline__contains:
                cmdline = " ".join(p.cmdline())
                if cmdline__contains not in cmdline:
                    continue
            # normal filters
            for k, v in kwargs.items():
                tmp = getattr(p, k)
                if callable(tmp):
                    tmp = tmp()
                if v != tmp:
                    break
            else:
                survivors.append(p)
        except (psutil.AccessDenied,):
            continue
        except (psutil.NoSuchProcess,):
            continue
    return survivors



""" pynchon.plugins.python.models

    Umbrella for different python-related plugins
"""

from .api import PythonAPI  # noqa
from .cli import PythonCLI  # noqa
from .platform import PythonPlatform  # noqa
from .pypi import PyPI  # noqa



""" pynchon.util.testing
"""

from pathlib import Path

from pynchon.abcs import AttrDict


def get_test_info(fname: str) -> dict:
    """

    :param fname: str:
    :param fname: str:

    """
    suite_dir = Path(fname).relative_to(Path.cwd()).parents[0]
    test_root = suite_dir.parents[0]
    return AttrDict(
        root=test_root,
        suite=AttrDict(path=suite_dir, name=suite_dir.name),
        fixtures=AttrDict(path=test_root / "fixtures"),
    )



""" pynchon.plugins.python.pypi
"""

from pynchon import abcs, models
from pynchon.util import lme, typing

LOGGER = lme.get_logger(__name__)


class PyPiConfig(abcs.Config):
    config_key: typing.ClassVar[str] = "pypi"
    name: str = typing.Field(default="Public PyPI")
    docs_url: str = typing.Field(default="https://pypi.org/")
    base_url: str = typing.Field(default="https://pypi.org/project")
    project_url: str = typing.Field(default=None)

    @property
    def project_url(self):
        from pynchon import config

        pname = config.project["name"]
        return f"{self.base_url}/{pname}"


class PyPI(models.Provider):
    """Context for PyPI"""

    name = "pypi"
    config_class = PyPiConfig



""" pynchon.models.plugins.provider """

from fleks import tagging

from . import validators

from pynchon import api, cli, events  # noqa
from pynchon.util import lme, typing  # noqa

from .cli import CliPlugin  # noqa

LOGGER = lme.get_logger(__name__)


@tagging.tags(cli_label="Provider")
class Provider(CliPlugin):
    """ProviderPlugin provides context-information,
    but little other functionality


    """

    cli_label = "Provider"
    contribute_plan_apply = False
    priority = 2
    __class_validators__ = [
        validators.require_conf_key,
        # validators.warn_config_kls,
    ]



""" pynchon.abcs.path
"""

import os
from fnmatch import fnmatch

from fleks.models import JSONEncoder

from pynchon.util import lme, typing

LOGGER = lme.get_logger(__name__)


class Path(typing.PathType):
    """ """

    def write(self, content: str) -> None:
        """ """
        with open(str(self), "w") as fhandle:
            fhandle.write(content)

    def read(self) -> str:
        """ """
        if not self.exists():
            raise ValueError(f"Cannot read nonexistant file @ {str(self)}")
        # LOGGER.critical(f'opening "{self}" for read..')
        with open(str(self)) as fhandle:
            content = fhandle.read()
        return content

    def endswith(self, other) -> bool:
        return str(self).endswith(other)

    def startswith(self, other) -> bool:
        return str(self).startswith(other)

    def path_truncated(self):
        """ """
        return self.parents[0] / self.stem_truncated()

    def full_extension(self):
        """no extension truncation, i.e. `.tar.gz`"""
        return self.name[self.name.find(".") :]

    def stem_truncated(self):
        """truncated stem"""
        return self.name[: self.name.rfind(self.full_extension())]

    def siblings(self):
        return self.parents[0].list()

    def match_any_glob(self, exclude_patterns: typing.List[str], quiet: bool = True):
        for exclude in exclude_patterns:
            match = self.match_glob(exclude)
            if match:
                quiet or LOGGER.info(f"{self} matches exclude @{match}")
                return match

    def match_glob(self, pattern):
        """

        :param pattern:

        """
        return fnmatch(str(self), str(pattern)) and pattern

    def has_file(self, fname) -> bool:
        """

        :param fname:

        """
        return self.absolute() in [p.absolute() for p in Path(fname).parents]

    def list(self) -> typing.List[str]:
        return [x for x in os.listdir(str(self))]


JSONEncoder.register_encoder(type=Path, fxn=str)



""" pynchon.plugins.python.libcst
"""

from fleks import cli, tagging

from pynchon.util.os import invoke

from pynchon import abcs, events, models  # noqa
from pynchon.util import lme, python, typing  # noqa

LOGGER = lme.get_logger(__name__)
F_CODEMOD_YAML = ".libcst.codemod.yaml"


class LibCST(models.Planner):
    """
    Code-transforms via libcst[1]
    """

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "python-libcst"

    name = "python-libcst"
    cli_name = "python-libcst"
    cli_label = "Tool"

    @cli.click.group("gen")
    def gen(self):
        """Generates code for python modules, packages, etc"""

    @tagging.tags(click_aliases=["src.transform"])
    @cli.click.argument("transform_name", nargs=1)
    @cli.click.argument("src_root", default="", nargs=1)
    def run_transform(self, transform_name="docstrings.simple.module", src_root=""):
        """Runs the given libcst transform on {src.root}"""
        src_root = src_root or self[:"src.root":]
        return invoke(
            f"python -m libcst.tool codemod {transform_name} {src_root}", system=True
        )

    @tagging.tags(click_aliases=["src.list-transforms"])
    def list_transforms(self):
        """Lists known libcst transforms"""
        out = invoke("python -mlibcst.tool list", strict=True)
        out = out.stdout
        print(f"\n{out}")

    def _goal_libcst_refresh(self, libcst_config):
        """ """
        from pynchon.util import text

        min = text.to_json(libcst_config, minified=True)
        rsrc = F_CODEMOD_YAML
        cmd = f"printf '{min}' | python -mpynchon.util.text.dumpf yaml > {rsrc}"
        return self.goal(
            type="render",
            label="refresh libcst-config",
            resource=rsrc,
            command=cmd,
        )

    def plan(self):
        """Run plan for this plugin"""
        plan = super(self.__class__, self).plan()
        libcst_config = self[F_CODEMOD_YAML::{}]
        if libcst_config:
            plan.append(self._goal_libcst_refresh(libcst_config))
        [plan.append(g) for g in self.docstrings(should_plan=True).goals]
        return plan

    @gen.command
    @cli.options.ignore_private
    @cli.options.ignore_missing
    @cli.options.plan
    @cli.click.argument("SRC_ROOT", required=False, default=None)
    @cli.click.option(
        "--modules", help="create docstrings for modules", default=False, is_flag=True
    )
    @cli.click.option(
        "--functions",
        help="create docstrings for functions",
        default=False,
        is_flag=True,
    )
    @cli.click.option(
        "--methods", help="create docstrings for methods", default=False, is_flag=True
    )
    def docstrings(
        self,
        src_root=None,
        should_plan: bool = False,
        ignore_missing: bool = False,
        ignore_private: bool = True,
        modules: bool = True,
        functions: bool = True,
        methods: bool = True,  # noqa
        classes: bool = True,
    ):
        """Generates python docstrings"""
        src_root = src_root or self[:"src.root":]
        cmd = "docstrings.simple.function"
        plan = self.Plan()
        plan.append(
            self.goal(
                command=f"python -mlibcst.tool codemod {cmd} {src_root}",
                resource=src_root,
                type="codemod",
            )
        )
        if should_plan:
            LOGGER.critical(plan)
            return plan
        else:
            return self.apply(plan)



""" pynchon.util.os
"""

# from . import models

from shil import invoke

# def invoke(cmd: str, **kwargs):
#     """
#     Dependency-free replacement for the `invoke` module,
#     which fixes annoyances with subprocess.POpen and os.system.
#     """
#     invoc = models.Invocation(cmd=cmd, **kwargs)
#     result = invoc()
#     result.__rich__()
#     return result


def slurp_json(cmd: str, **kwargs):
    """ """
    result = invoke(f"{cmd} > .tmp.{id(cmd)}")
    assert result.succeeded
    from pynchon.util.text import loadf

    return loadf.json(f".tmp.{id(cmd)}")



""" pynchon.plugins.util
"""

from pynchon.util import lme, typing  # noqa

# , importing

# from .exceptions import * # noqa
LOGGER = lme.get_logger(__name__)


# importing.module_builder(__name__, ...)
class PluginNotInitialized(RuntimeError):
    pass


class PluginNotRegistered(RuntimeError):
    pass


class PluginNotConfigured(RuntimeError):
    pass


def get_plugin_meta(plugin_name: str, strict: bool = True) -> typing.Dict:
    """ """
    from pynchon.plugins import registry

    try:
        return registry[plugin_name]
    except KeyError:
        if strict:
            raise PluginNotRegistered(plugin_name)
        else:
            return None


def get_plugin_class(plugin_name: str, strict: bool = True) -> typing.Type:
    """ """
    meta = get_plugin_meta(plugin_name, strict=strict)
    if meta is None:
        return meta
    try:
        return meta["kls"]
    except KeyError:
        if strict:
            raise PluginNotRegistered(plugin_name)
        else:
            return None


get_plugin = get_plugin_class


def get_plugin_obj(plugin_name: str) -> object:
    """

    :param plugin_name: str:
    :param plugin_name: str:

    """
    meta = get_plugin_meta(plugin_name)
    try:
        return meta["obj"]
    except KeyError:
        err = f"cannot retrieve ['obj'] for `{plugin_name}` from registry; is config finalized?"
        LOGGER.critical(err)
        raise PluginNotInitialized(plugin_name)



"""
"""

import webbrowser

from pynchon import abcs
from pynchon.util import lme

LOGGER = lme.get_logger(__name__)


class OpenerMixin:
    """
    Helper for opening project documentation-files
    inside a webbrowser.  We use a modified version
    of `grip`[1] for this which is called `gripe`[2].

    How this actually works depends on the file-types
    involved, because whereas `grip` works natively to
    render markdown as github-flavored markdown, it
    ignores other types by default.  The `gripe` tool
    adds a special /__raw__ endpoint for other kinds of
    static content.

    (In the future other file-types like `.dot` for raw
    graphviz might be supported, but for now you'll still
    have to render those to png to see a picture.)
    """

    def _open_grip(self, file: str = None):
        """

        :param file: str:  (Default value = None)
        :param file: str:  (Default value = None)

        """
        pfile = abcs.Path(file).absolute()
        relf = pfile.relative_to(abcs.Path(self.git_root))
        port = self.server.port
        if port is None:
            LOGGER.critical("no server yet..")
            self.serve()
            import time

            time.sleep(3)
            return self._open_grip(file=file)
        grip_url = f"http://localhost:{port}/{relf}"
        LOGGER.warning(f"opening {grip_url}")
        return dict(url=grip_url, browser=webbrowser.open(grip_url))

    _open__md = _open_grip
    _open__mmd = _open_grip

    def _open_raw(self, file: str = None, server=None):
        """
        :param file: str:  (Default value = None)
        :param server: Default value = None)
        """
        relf = file.absolute().relative_to(abcs.Path(self.git_root))
        return self._open_grip(abcs.Path("__raw__") / relf)

    _open__html = _open_raw
    _open__png = _open_raw
    _open__jpg = _open_raw
    _open__jpeg = _open_raw
    _open__gif = _open_raw
    _open__svg = _open_raw
    _open__htm = _open__html



"""
pynchon: a utility for docs generation and template-rendering
"""

import fleks
import shimport

from pynchon import cli

from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)

click = cli.click
plugins = shimport.lazy("pynchon.plugins")

from trogon import tui


class RootGroup(fleks.cli.RootGroup):
    @fleks.classproperty
    def default(kls):  # noqa
        from pynchon import bin

        return bin.default


@tui()
@click.version_option()
@click.option("--plugins", help="shortcut for `--set plugins=...`")
@click.option("--set", "set_config", help="config overrides")
@click.option("--get", "get_config", help="config retrieval")
@click.group(
    "pynchon",
    cls=RootGroup,
)
def entry(
    plugins: str = "",
    set_config: str = "",  # noqa
    get_config: str = "",
):
    """ """


def bootstrap():
    """
    :param :
    """
    from pynchon.app import app
    from pynchon.plugins import registry as plugin_registry

    from pynchon import config  # isort: skip

    events = app.events
    events.lifecycle.send(__name__, stage="Building CLIs from plugins..")
    registry = click_registry = {}
    loop = plugin_registry.items()
    for name, plugin_meta in loop:
        if name not in config.PLUGINS:
            LOGGER.warning(f"skipping `{name}`")
            continue
        plugin_kls = plugin_meta["kls"]
        init_fxn = plugin_kls.init_cli
        # LOGGER.critical(f'\t{name}.init_cli: {init_fxn}')
        try:
            p_entry = init_fxn()
        except (Exception,) as exc:
            LOGGER.critical(f"  failed to initialize cli for {plugin_kls.__name__}:")
            LOGGER.critical(f"    {exc}")
            raise
        else:
            registry[name] = dict(plugin=plugin_kls, entry=p_entry)


@entry.command(
    "default",
    hidden=True,
    context_settings=dict(
        ignore_unknown_options=True,
    ),
)
@click.option("--plugins", help="shortcut for `--set plugins=...`")
@click.option("--set", "set_config", help="config overrides")
@click.option("--get", "get_config", help="config retrieval")
@click.argument("extra", nargs=-1)
@click.pass_context
def default(
    ctx, plugins: str = "", set_config: str = "", get_config: str = "", **kwargs  # noqa
):
    """this is always executed, regardless of subcommands and before them"""
    # LOGGER.critical('top-level')
    setters = ctx.params.get("set_config", []) or []
    plugins = ctx.params.get("plugins", "")
    plugins and setters.append([f'pynchon.plugins={plugins.split(",")}'])
    setters and LOGGER.critical(f"--set: {setters}")
    bootstrap()



""" pynchon.plugins.deck
"""

from pynchon import abcs, cli, events, models  # noqa
from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


class Deck(models.ResourceManager):
    """Tool for working with markdown based slide-decks"""

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "deck"
        root: str = typing.Field(default="{{docs.root}}/slides")
        pandoc_docker: str = typing.Field(default="pandoc/core")
        pandoc_engine: str = typing.Field(default="dzslides")
        pandoc_args: typing.List[str] = typing.Field(default=[])
        apply_hooks: typing.List[str] = typing.Field(default=["open-after"])
        include_patterns: typing.List[str] = typing.Field(default=["*.md"])

    name = "deck"
    cli_name = "deck"
    cli_label = "Tool"
    contribute_plan_apply = True

    def plan(self, **kwargs):
        plan = super().plan()
        root = self["root"]
        root = abcs.Path(root)
        if not root.exists():
            plan.append(
                self.goal(resource=root, type="mkdir", command=f"mkdir -p {root}")
            )
        for rsrc in self.list():
            output = rsrc.parents[0] / (rsrc.stem + ".html")
            proot = self[:"pynchon.root"]
            output = output.relative_to(proot)
            relr = rsrc.relative_to(proot)
            fargs = {
                **self.config.dict(),
                **dict(
                    relr=relr, pandoc_args=" ".join(self["pandoc_args"]), output=output
                ),
            }
            plan.append(
                self.goal(
                    resource=output.absolute(),
                    type="gen",
                    command=(
                        "docker run -v `pwd`:/workspace "
                        "-w /workspace "
                        "{pandoc_docker} "
                        "-t {pandoc_engine} -s {relr} -o {output} {pandoc_args}"
                    ).format(**fargs),
                )
            )
        return plan



""" pynchon.codemod.docstrings.base """

import argparse

from libcst.codemod import CodemodContext

from pynchon.util import lme

LOGGER = lme.get_logger(__name__)

from libcst.codemod import ContextAwareTransformer


class base(ContextAwareTransformer):  # VisitorBasedCodemodCommand):
    DESCRIPTION: str = """\n\tAbstract, don't use this directly"""

    def __init__(
        self,
        context: CodemodContext,
        # formatter: str = "black",
        # parameter_count: Optional[int] = None,
        # argument_count: Optional[int] = None,
    ) -> None:
        super().__init__(context)
        # self.parameter_count: int = parameter_count or presets["parameter_count"]
        # self.argument_count: int = argument_count or presets["argument_count"]

    @staticmethod
    def add_args(arg_parser: argparse.ArgumentParser) -> None:  # noqa
        """ """
        # arg_parser.add_argument(
        #     "--formatter",
        #     dest="formatter",
        #     metavar="FORMATTER",
        #     help="Formatter to target (e.g. yapf or black)",
        #     type=str,
        #     default="black",
        # )
        # arg_parser.add_argument(
        #     "--paramter-count",
        #     dest="parameter_count",
        #     metavar="PARAMETER_COUNT",
        #     help="Minimal number of parameters for us to add trailing comma",
        #     type=int,
        #     default=None,
        # )
        # arg_parser.add_argument(
        #     "--argument-count",
        #     dest="argument_count",
        #     metavar="ARGUMENT_COUNT",
        #     help="Minimal number of arguments for us to add trailing comma",
        #     type=int,
        #     default=None,
        # )

    # def leave_Parameters(
    #     self,
    #     original_node: cst.Parameters,
    #     updated_node: cst.Parameters,
    # ) -> cst.Parameters:
    # skip = (
    #     #
    #     self.parameter_count is None
    #     or len(updated_node.params) < self.parameter_count
    #     or (
    #         len(updated_node.params) == 1
    #         and updated_node.params[0].name.value in {"self", "cls"}
    #     )
    # )
    # if skip:
    #     return updated_node
    # else:
    #     last_param = updated_node.params[-1]
    #     return updated_node.with_changes(
    #         params=(
    #             *updated_node.params[:-1],
    #             last_param.with_changes(comma=cst.Comma()),
    #         ),
    #     )

    # def leave_SimpleString(
    #     self, onode, unode
    # ):
    #     import IPython; IPython.embed()



""" pynchon.config.util
"""

import functools
import collections

from pynchon import abcs
from pynchon import config as config_module
from pynchon import constants, events
from pynchon.util import lme, typing
from pynchon.util.text import loadf, loads
from pynchon.plugins.util import PluginNotRegistered, get_plugin

LOGGER = lme.get_logger(__name__)

from pydantic import create_model


@functools.lru_cache(maxsize=100, typed=False)
def finalize():
    """ """
    from pynchon import abcs
    from pynchon.plugins import registry as plugins_registry

    plugins = []
    top = dict(
        __base__=abcs.Config,
        pynchon=config_module.RAW,
        git=config_module.GIT,
    )

    # NB: already sorted by priority
    for pname in plugins_registry.keys():
        try:
            tmp = get_plugin(pname)
        except (PluginNotRegistered,) as exc:
            LOGGER.critical(f"PluginNotRegistered: {exc}")
            continue
        else:
            plugins.append(tmp)

    for plugin_kls in plugins:
        pconf_kls = getattr(plugin_kls, "config_class", None)
        conf_key = plugin_kls.get_config_key()
        if pconf_kls is None:
            plugin_config = abcs.Config()
        else:
            # user_defaults = (
            #     config_module.PYNCHON_CORE
            #     if plugin_kls.name == "base"
            #     else config_module.USER_DEFAULTS.get(plugin_kls.name, {})
            # )
            # user_defaults = config_module.USER_DEFAULTS.get(plugin_kls.name, {})
            user_defaults = config_module.MERGED_CONFIG_FILES.get(conf_key, {})
            ctx = {
                **config_module.MERGED_CONFIG_FILES,
                **dict(pynchon=config_module.PYNCHON),
            }
            # if conf_key=='globals':
            #     import IPython; IPython.embed()
            from pynchon.api import render

            user_defaults = render.dictionary(user_defaults, ctx)

            if plugin_kls.name == "core":
                # special case: this is already bootstrapped
                from pynchon.config import PYNCHON_CORE

                plugin_config = PYNCHON_CORE
            else:
                plugin_config = pconf_kls(
                    **{
                        # **plugin_defaults,
                        **user_defaults,
                    }
                )
        setattr(config_module, conf_key, plugin_config)
        # result.update({conf_key: plugin_config})
        if conf_key not in ["json"]:
            top[conf_key] = plugin_config
        else:
            top[f"{conf_key}_"] = typing.Field(alias=conf_key, default=plugin_config)
        events.lifecycle.send(
            __name__, config=f"plugin-config@`{plugin_config}` was finalized"
        )

        plugin_obj = plugin_kls(final=plugin_config)
        # plugins_registry.register(plugin_obj)
        plugins_registry[plugin_kls.name]["obj"] = plugin_obj
        events.lifecycle.send(
            plugin_obj, plugin=f"plugin@`{plugin_kls.__name__}` was finalized"
        )
    result = create_model("Top", **top)
    result = result()
    # import IPython; IPython.embed()
    return result


def config_folders():
    import os

    from pynchon import config

    folders = list(
        set(
            filter(
                None,
                [
                    os.getcwd(),
                    constants.PYNCHON_ROOT,
                    config.GIT["root"],
                ],
            )
        )
    )
    return [abcs.Path(f) for f in folders]


def get_config_files():
    """ """
    if constants.PYNCHON_CONFIG:
        return [abcs.Path(constants.PYNCHON_CONFIG)]
    result = []
    folders = config_folders()
    for folder in folders:
        for file in constants.CONF_FILE_SEARCH_ORDER:
            candidate = folder / file
            result.append(candidate)
    return result


def load_config_from_files() -> typing.Dict[str, str]:
    """ """
    contents = collections.OrderedDict()
    candidates = get_config_files()
    for config_file in candidates:
        if not config_file.exists():
            LOGGER.info(f"config_file@`{config_file}` doesn't exist, skipping it")
            continue
        elif config_file.name.endswith("pyproject.toml"):
            LOGGER.info(f"Loading from toml: {config_file}")
            tmp = loadf.toml(config_file)
            tmp = tmp.get("tool", {})
            tmp = tmp.get("pynchon", {})
            contents[config_file] = tmp
        elif config_file.name.endswith(".json5"):
            LOGGER.info(f"Loading from json5: {config_file}")
            with open(config_file.absolute()) as fhandle:
                raw_contents = fhandle.read()
            try:
                contents[config_file] = loads.json5(raw_contents)
            except (Exception,) as exc:
                LOGGER.critical(exc)
                contents[config_file] = {}

        else:
            err = f"don't know how to load config from {config_file}"
            raise NotImplementedError(err)
    return contents



""" pynchon.plugins.plugins
"""

from fleks import tagging

from pynchon import abcs, cli, models
from pynchon.util.os import invoke

from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


class PluginsMan(models.Manager):
    """Meta-plugin for managing plugins"""

    name = "plugins"
    cli_name = "plugins"

    @cli.click.option("--name")
    @cli.click.option("--template-skeleton", "-t", is_flag=True, default=False)
    def new(self, name: str = None, template_skeleton: bool = False) -> None:
        """Create new plugin from template (for devs)

        :param name: str:  (Default value = None)
        :param template_skeleton: bool:  (Default value = False)
        """
        # FIXME: use pattern?
        plugins_d = abcs.Path(__file__).parents[0]
        template_plugin_f = plugins_d / "__template__.py"
        new_plugin_file = plugins_d / f"{name}.py"
        cmd = f"ls {new_plugin_file} || cp {template_plugin_f} {new_plugin_file} && git status"
        result = invoke(cmd, system=True)
        if template_skeleton:
            raise NotImplementedError()
        return result.succeeded

    @tagging.tags(click_aliases=["ls"])
    def list(self, **kwargs):
        """List all plugins"""
        return list(self.status()["plugins"].keys())

    @tagging.tags(click_aliases=["st", "stat"])
    def status(self) -> typing.Dict:
        """Returns details about all known plugins"""
        result = typing.OrderedDict()
        for name, p in self.siblings.items():
            result[name] = dict(priority=p.priority, key=p.get_config_key())
        return dict(plugins=result)



""" pynchon.plugins.gen
"""

from pynchon import abcs, cli, models  # noqa
from pynchon.util import lme, typing

LOGGER = lme.get_logger(__name__)


class Generators(models.NameSpace):
    """Collects `gen` commands from other plugins"""

    name = cli_name = "gen"
    priority = 1
    config_class = None

    @classmethod
    def siblings_with_subcommand(
        kls, target_name: str, siblings: typing.Dict = {}
    ) -> typing.Dict[object, typing.Callable]:
        result = {}
        for name, obj in siblings.items():
            if obj.name == "core":
                continue
            elif target_name in dir(obj):
                result[obj] = getattr(obj, target_name)
        return result

    @classmethod
    def acquire_cli_subcommands(kls, target_name, **kwargs) -> None:
        def decide_subcommand_name(apparent_name):
            cmd_name = apparent_name
            if cmd_name in [target_name]:
                cmd_name = sibling.name
            if cmd_name in cmd_names:
                cmd_name = f"{sibling.name}-{cmd_name}"
            cmd_name = cmd_name.replace("_", "-")
            if cmd_name in ["toc", "detail"]:
                cmd_name = "-".join([cmd_name, sibling.name.split("-")[-1]])
            if cmd_name in cmd_names:
                raise Exception(f"collision acquiring subcommand: [{cmd_name}, {cmd}]")
            return cmd_name

        def get_cli_path(sibling, group, ocmd_name):
            cmd_name = decide_subcommand_name(ocmd_name)
            gname = f"{group.name} " if group else ""
            cname = f"{gname}{cmd_name}"
            ocli = f"`{sibling.cli_path} {gname}{ocmd_name}`"
            return ocli

        def get_name(cmd):
            return getattr(cmd, "name", getattr(cmd, "__name__", ""))

        def acqcmd(cmd, group=None):
            return kls.click_acquire(
                cmd,
                copy=True,
                name=decide_subcommand_name(get_name(cmd)),
                help=" ".join(
                    [
                        (
                            "Aliases from"
                            if isinstance(cmd, (cli.click.Group,))
                            else "Alias to"
                        ),
                        get_cli_path(sibling, group, get_name(cmd)),
                    ]
                ),
            )

        cmd_names = []
        for sibling, cmd in kls.siblings_with_subcommand("gen", **kwargs).items():
            LOGGER.info(f"acquiring {cmd} (type={type(cmd)}) from {sibling}")
            if isinstance(cmd, cli.click.Group):
                [acqcmd(c, group=cmd) for c in cmd.commands.values()]
            else:
                acqcmd(cmd)

    @classmethod
    def init_cli(kls) -> cli.click.Group:
        """ """
        result = super(kls, kls).init_cli()
        kls.acquire_cli_subcommands("gen", siblings=kls.siblings)
        return result



""" pynchon.util.text.loadf

    Helpers for loading data structures from files
"""

import os

import tomli as tomllib  # NB: tomllib only available in py3.11
from fleks.cli import click, options

from pynchon.util.os import invoke
from pynchon.util.text import loads

from pynchon.util import lme, text, typing  # noqa

LOGGER = lme.get_logger(__name__)


def loadf(file=None, content=None):
    """ """
    if file:
        assert not content
        if not os.path.exists(file):
            raise ValueError(f"file @ {file} is missing!")
        with open(file) as fhandle:
            content = fhandle.read()
    return content


@click.argument("file", nargs=1)
def ini(file):
    """Parses ini file and returns JSON

    :param file:

    """
    import configparser

    parser = configparser.ConfigParser()
    parser.read(file)
    ini_conf = {section: dict(parser.items(section)) for section in parser.sections()}
    return ini_conf


def yaml(fname: str) -> typing.Dict:
    """parses yaml file and returns JSON
    :param *args:
    :param **kwargs:
    """
    from pynchon.util.text import loads

    with open(fname) as fhandle:
        contents = fhandle.read()
    return loads.yaml(contents)


@click.argument("file", nargs=1)
def toml(file: str = None, strict: bool = True):
    """Parses toml file and returns JSON

    :param file: str:  (Default value = None)
    :param strict: bool:  (Default value = True)
    """

    config = {}
    if not os.path.exists(file):
        err = f"Cannot load config from nonexistent path @ `{file}`"
        LOGGER.critical(err)
        if strict:
            raise
        else:
            return config
    with open(file, "rb") as f:
        try:
            config = tomllib.load(f)
        except (tomllib.TOMLDecodeError,) as exc:
            LOGGER.critical(f"cannot decode data from toml @ {file}")
            raise
    return config


# @click.argument("file", nargs=1)
# def json5(file: str = '',) -> dict:
#     """
#     loads json5 from file
#     """
#     assert file
#     if not os.path.exists(file):
#         raise ValueError(f'file @ {file} is missing!')
#     with open(file, 'r') as fhandle:
#         content = fhandle.read()
#     data = loads.json5(content)
#     return data


@click.option(
    "--wrap-with-key",
    "wrapper_key",
    help="when set, wraps output as `{WRAPPER_KEY:output}`",
    default="",
)
@options.output
@options.should_print
@click.option("--pull", help="when provided, this key will be output", default="")
@click.option(
    "--push-data", help="(string) this raw data will be added to output", default=""
)
@click.option(
    "--push-file-data",
    help="(filename) contents of file will be added to output",
    default="",
)
@click.option(
    "--push-json-data",
    help="(string) jsonified data will be added to output",
    default="",
)
@click.option(
    "--push-command-output", help="command's stdout will be added to output", default=""
)
@click.option("--under-key", help="required with --push commands", default="")
@click.argument("files", nargs=-1)
def json5(
    file: str = "",
    files: typing.List[str] = [],
    output: str = "",
    should_print: bool = False,
    wrapper_key: str = "",
    pull: str = "",
    push_data: str = "",
    push_file_data: str = "",
    push_json_data: str = "",
    push_command_output: str = "",
    under_key: str = "",
) -> None:
    """Parses JSON-5 file(s) and outputs json. Pipe friendly.

    If multiple files are provided, files will
    be merged (with overwrites) in the order provided.

    Several other options are available for common post-processing tasks.

    :param output: str:  (Default value = '')
    :param should_print: bool:  (Default value = False)
    :param file: str:  (Default value = '')
    :param files: typing.List[str]:  (Default value = [])
    :param wrapper_key: str:  (Default value = '')
    :param pull: str:  (Default value = '')
    :param push_data: str:  (Default value = '')
    :param push_file_data: str:  (Default value = '')
    :param push_json_data: str:  (Default value = '')
    :param push_command_output: str:  (Default value = '')
    :param under_key: str:  (Default value = '')
    :param output: str:  (Default value = '')
    :param should_print: bool:  (Default value = False)
    :param file: str:  (Default value = '')
    :param files: typing.List[str]:  (Default value = [])
    :param wrapper_key: str:  (Default value = '')
    :param pull: str:  (Default value = '')
    :param push_data: str:  (Default value = '')
    :param push_file_data: str:  (Default value = '')
    :param push_json_data: str:  (Default value = '')
    :param push_command_output: str:  (Default value = '')
    :param under_key: str:  (Default value = '')

    """
    out: typing.Dict[str, typing.Any] = {}
    files = files or (file and [file]) or []
    for file in files:
        with open(file) as fhandle:
            obj = loads.json5(fhandle.read())
        out = {**out, **obj}

    push_args = [push_data, push_file_data, push_json_data, push_command_output]
    if any(push_args):
        assert under_key
        assert under_key not in out, f"content already has key@{under_key}!"
        assert (
            sum([1 for x in push_args if x]) == 1
        ), "only one --push arg can be provided!"
        if push_data:
            assert isinstance(push_data, (str,))
            push = push_data
        elif push_command_output:
            cmd = invoke(push_command_output)
            if cmd.succeeded:
                push = cmd.stdout
            else:
                err = cmd.stderr
                LOGGER.critical(err)
                raise SystemExit(1)
        elif push_json_data:
            push = loads.json5(content=push_json_data)
        elif push_file_data:
            err = f"file@{push_file_data} doesnt exist"
            assert os.path.exists(push_file_data), err
            with open(push_file_data) as fhandle:
                push = fhandle.read()
        out[under_key] = push

    if wrapper_key:
        # NB: must remain after push
        out = {wrapper_key: out}

    if pull:
        out = out[pull]
        # similar to `jq -r`.
        # we don't want quoted strings, but
        # if the value is complex, we need json-encoding
        if not isinstance(out, (str,)):
            msg = text.to_json(out)
        else:
            msg = str(out)
    else:
        msg = text.to_json(out)
    output = output or "/dev/stdout"
    print(msg, file=open(output, "w"))
    if should_print and output != "/dev/stdout":
        print(msg)
    return out


@options.strict
@click.argument("file", nargs=1)
def json(file: str = "", content: str = "", strict: bool = True) -> dict:
    """loads json to python dictionary from given file or string

    :param file: str:  (Default value = '')
    :param content: str:  (Default value = '')
    :param strict: bool:  (Default value = True)
    :param file: str:  (Default value = '')
    :param content: str:  (Default value = '')
    :param strict: bool:  (Default value = True)

    """
    content = content or text.loadf.loadf(file=file, content=content)
    try:
        data = loads.json(content)
        # data = pyjson5.loads(content)
    # except (pyjson5.Json5EOF,) as exc:
    except (ValueError,) as exc:
        LOGGER.critical(f"Cannot parse json from {file}!")
        raise
    return data



""" pynchon.plugins.project
"""

from pynchon import abcs, config, constants, models

from pynchon.cli import common, options  # noqa
from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


class ProjectConfig(abcs.Config):
    """ """

    config_key: typing.ClassVar[str] = "project"
    shell_aliases: typing.Dict[str, str] = typing.Field(default={})
    subproject_patterns: typing.List[str] = typing.Field(default=[])

    @property
    def name(self) -> typing.StringMaybe:
        """ """
        repo_name = config.git.repo_name
        return repo_name or abcs.Path(".").name

    @property
    def _workdir(self):
        return abcs.Path(".").absolute()

    @property
    def root(self) -> str:
        """ """
        git = config.GIT
        return constants.PYNCHON_ROOT or (git and git.root) or self._workdir

    @property
    def subproject(self) -> typing.Dict:
        """ """
        if constants.PYNCHON_ROOT:
            return {}
        git = config.GIT
        git_root = git["root"]
        r1 = self._workdir
        r2 = git_root and git_root.absolute()
        if r2 and (r1 != r2):
            LOGGER.warning(f"subproject detected:\n\t({r1} != git[root] @ {r2})")
            return dict(name=self._workdir.name, root=self._workdir)
        return {}


class Project(models.Manager):
    """Meta-plugin for managing this project"""

    name = "project"
    priority = 2
    config_class = ProjectConfig

    # @common.kommand(
    #     name="version",
    #     parent=parent,
    #     formatters=dict(markdown=constants.T_VERSION_METADATA),
    #     options=[
    #         # FIXME: options.output_with_default('docs/VERSION.md'),
    #         options.format_markdown,
    #         options.output,
    #         options.header,
    #     ],
    # )
    # def project_version(format, output, header) -> None:
    #     """
    #     Describes version details for this package (and pynchon itself).
    #     """
    #     # from pynchon.api import python #, git
    #     import pynchon
    #     from pynchon.config import git, python
    #
    #     return dict(
    #         pynchon_version=pynchon.__version__,
    #         package_version=python.package.version,
    #         git_hash=git.hash,
    #     )



""" pynchon.plugins.mkdocs
"""

from pathlib import Path

from pynchon.plugins import util as plugin_util
from pynchon.util.text import loadf

from pynchon import abcs, api, events, models  # noqa
from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


class MkdocsPluginConfig(abcs.Config):
    config_key: typing.ClassVar[str] = "mkdocs"
    config_file: str = typing.Field(default=None)

    @property
    def site_dir(self):
        return self.config.get("site_dir", "site")

    @property
    def config(self) -> typing.Dict:
        """
        returns a dictionary with the current mkdocs configuration
        """
        fname = self.config_file
        if fname is None:
            return {}

        return loadf.yaml(fname)

    @property
    def config_file(self) -> typing.StringMaybe:
        """returns the path to the mkdocs config-file, if applicable"""
        docs = plugin_util.get_plugin("docs", strict=False)
        docs = docs and docs.get_current_config()
        subproject = plugin_util.get_plugin("subproject", strict=False)
        subproject = subproject and subproject.get_current_config()
        project = plugin_util.get_plugin("project", strict=False)
        project = project and project.get_current_config()
        candidates = filter(
            None,
            [
                abcs.Path(".").absolute(),
                docs and docs.root,
                subproject and subproject.root,
                project and project.root,
            ],
        )
        for folder in [Path(c) for c in candidates]:
            cand = folder / "mkdocs.yml"
            if cand.exists():
                return str(cand.absolute())


class Mkdocs(models.Planner):
    """Mkdocs helper"""

    priority = 6  # before mermaid
    name = "mkdocs"
    cli_name = "mkdocs"
    cli_label = "Docs"
    config_class = MkdocsPluginConfig

    def open(self):
        """
        Opens `site_dir` in a webbrowser
        """
        import webbrowser

        index_f = Path(self.site_dir).absolute() / "index.html"
        url = f"file://{index_f}"
        return webbrowser.open(url)

    @property
    def site_dir(self) -> str:
        """
        Returns mkdocs `site_dir` if present in config, or guesses what it should be
        """
        plugin_cfg = self.config
        mkdocs_config = plugin_cfg.config
        result = str(mkdocs_config.get("site_dir", self.working_dir / "site"))
        self.logger.warning(f"returning {result}")
        return result

    # def _hook_open_after_apply(self, result) -> bool:
    #     raise Exception(result)

    def plan(self):
        """
        Runs a plan for this plugin
        """
        plan = super(self.__class__, self).plan()
        config_file = self["config_file"]
        plan.append(
            self.goal(
                type="render",
                resource=self.site_dir,
                command=f"mkdocs build --config-file {config_file}",
            )
        )
        return plan



""" pynchon.models.validators

    FIXME: ntuple struct for vdata
"""

import fleks

from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


# @validator
def require_conf_key(
    kls,
    self=None,
    vdata=None,
    strict: bool = True,
):
    """ """
    pconf_kls = getattr(kls, "config_class", None)
    conf_key = getattr(pconf_kls, "config_key", kls.name.replace("-", "_"))
    if not conf_key:
        msg = f"failed to determine conf-key for {kls}"
        LOGGER.critical(msg)
        if strict:
            # raise vdata.error_class(msg)
            raise fleks.meta.ClassMalformed(msg)
    return vdata


def warn_config_kls(kls, self=None, vdata=None):
    """
    :param vdata=None:
    :param self=None:
    :param kls:
    """
    """ """
    """ """
    pconf_kls = getattr(kls, "config_class", "NOTSET")
    if pconf_kls == "NOTSET":
        vdata.warnings["`config_class` not set!"].append(kls)
    return vdata



""" pynchon.util.makefile
"""

import os
import re

from pynchon import abcs, cli
from pynchon.util.os import invoke

from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)
zzz = "#  recipe to execute (from '"
vvv = "# Variables"
fff = "# files hash-table stats:"


@cli.click.argument("makefile")
def database(makefile: str = "", make="make") -> typing.List[str]:
    """
    Get database for Makefile (i.e. 'make --print-data-base')
    """
    assert makefile
    tmp = abcs.Path(makefile)
    if not all(
        [
            tmp.exists,
            tmp.is_file,
        ]
    ):
        raise ValueError(f"{makefile} does not exist")
    else:
        LOGGER.warning(f"parsing makefile @ {makefile}")
    cmd = f"{make} --print-data-base -pqRrs -f {makefile}"
    resp = invoke(cmd)
    out = resp.stdout.split("\n")
    return out


def _test(x):
    """ """
    return all(
        [
            ":" in x.strip(),
            not x.startswith("#"),
            not x.startswith("."),
            not x.startswith("\t"),
        ]
    )


def _get_prov_line(body):
    # zzz = "#  recipe to execute (from '"
    pline = [x for x in body if zzz in x]
    pline = pline[0] if pline else None
    return pline


def _get_file(body=None, makefile=None):
    pline = _get_prov_line(body)
    if pline:
        return pline.split(zzz)[-1].split("'")[0]
    else:
        return str(makefile)


@cli.click.argument("makefile")
def parse(makefile: str = None, bodies=False, **kwargs):
    """
    Parse Makefile to JSON.  Includes targets/prereq detail
    """
    assert os.path.exists(makefile)
    wd = abcs.Path(".")
    db = database(makefile, **kwargs)
    original = open(makefile).readlines()
    variables_start = db.index(vvv)
    variables_end = db.index("", variables_start + 2)
    vars = db[variables_start:variables_end]
    db = db[variables_end:]
    implicit_rule_start = db.index("# Implicit Rules")
    file_rule_start = db.index("# Files")
    file_rule_end = db.index(fff)
    for i, line in enumerate(db[implicit_rule_start:]):
        if "implicit rules, " in line and line.endswith(" terminal."):
            implicit_rule_end = implicit_rule_start + i
            break
    else:
        LOGGER.critical("cannot find `implicit_rule_end`!")
        implicit_rule_end = implicit_rule_start
    implicit_targets_section = db[implicit_rule_start:implicit_rule_end]
    file_targets_section = db[file_rule_start:file_rule_end]
    file_target_names = list(filter(_test, file_targets_section))
    implicit_target_names = list(filter(_test, implicit_targets_section))
    targets = file_target_names + implicit_target_names
    out = {}
    for tline in targets:
        bits = tline.split(":")
        target_name = bits.pop(0)
        childs = ":".join(bits)
        type = "implicit" if tline in implicit_targets_section else "file"
        # NB: line nos are from reformatted output, not original file
        line_start = db.index(tline)
        line_end = db.index("", line_start)
        body = db[line_start:line_end]
        pline = _get_prov_line(body)
        file = _get_file(body=body, makefile=makefile)
        if pline:
            # take advice from make's database.
            # we return this because it's authoritative,
            # but actually sometimes it's wrong.  this returns
            # the first like of the target that's tab-indented,
            # but sometimes make macros like `ifeq` are not indented..
            lineno = pline.split("', line ")[-1].split("):")[0]
        else:
            try:
                lineno = original.index(tline)
            except ValueError:
                LOGGER.critical(f"cant find {tline} in {file}, parametric?")
                # target_name
                lineno = None
        lineno = lineno and (int(lineno) - 1)
        prereqs = [x for x in childs.split() if x.strip()]
        out[target_name] = dict(
            file=file,
            lineno=lineno,
            body=body,
            type=type,
            docs=[x[len("\t@#") :] for x in body if x.startswith("\t@#")],
            prereqs=prereqs,
        )
        if type == "implicit":
            regex = target_name.replace("%", ".*")
            out[target_name].update(regex=regex)
    for target_name, tmeta in out.items():
        if "regex" in tmeta:
            implementors = []
            for impl in out:
                if impl != target_name and re.compile(tmeta["regex"]).match(impl):
                    implementors.append(impl)
            tmeta["implementors"] = implementors

    for target_name, tmeta in out.items():
        real_body = [
            b
            for b in tmeta["body"][1:]
            if not b.startswith("#") and not b.startswith("@#")
        ]
        if not real_body:
            LOGGER.critical(f"missing body for: {target_name}")
            for chain in out:
                if target_name in out[chain].get("implementors", []):
                    tmeta["chain"] = chain
            if len(tmeta["prereqs"]) == 1:
                tmeta["chain"] = tmeta["prereqs"][0]

    if not bodies:
        tmp = {}
        for k, v in out.items():
            v.pop("body", [])
            tmp[k] = v
        out = tmp
    return out


parser = parse



""" pynchon.util
"""

from pynchon import constants

from . import lme

LOGGER = lme.get_logger(__name__)


def get_root(path: str = ".") -> str:
    """

    :param path: str:  (Default value = ".")
    :param path: str:  (Default value = ".")

    """
    import os

    from pynchon.abcs import Path

    path = Path(path).absolute()
    if (path / ".git").exists():
        return path.relative_to(os.getcwd())
    elif not path:
        return None
    else:
        return get_root(path.parents[0])


def is_python_project() -> bool:
    """ """
    import os

    from pynchon.api import git

    return os.path.exists(os.path.join(git.get_root(), constants.PYNCHON_CONFIG_FILE))


def find_src_root(config: dict) -> str:
    """

    :param config: dict:
    :param config: dict:

    """
    from pynchon.abcs import Path

    pconf = config.get("project", {})
    LOGGER.debug(f"project config: {pconf}")
    src_root = Path(pconf.get("src_root", "."))
    src_root = src_root if src_root.is_dir() else None
    return src_root.relative_to(".")



""" pynchon.util.files CLI
"""

import shimport
from fleks.util.tagging import tags

from pynchon.cli import common

from fleks.cli import click, options  # noqa

from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


entry = common.entry_for(__name__)


(
    shimport.wrapper("pynchon.util.files")
    .prune(
        filter_instances=typing.FunctionType,
        filter_module_origin="pynchon.util.files",
    )
    .map_ns(
        lambda _name, fxn: [fxn, tags[fxn].get("click_aliases", []) + [fxn.__name__]]
    )
    .starmap(
        lambda fxn, aliases: [
            entry.command(
                name=alias.replace("_", "-"),
                help=(
                    fxn.__doc__
                    if alias == fxn.__name__
                    else f"alias for `{fxn.__name__}`"
                ),
            )(fxn)
            for alias in aliases
        ]
    )
)
if __name__ == "__main__":
    entry()



""" pynchon.models.planner """

import typing

from fleks import tagging
from memoized_property import memoized_property
from fleks.util.tagging import tags

from pynchon import abcs, cli
from pynchon.app import app
from pynchon.util.os import invoke

from . import planning
from .plugins import BasePlugin

from pynchon.util import lme, typing  # noqa


LOGGER = lme.get_logger(__name__)


@tags(cli_label="Planner")
class AbstractPlanner(BasePlugin):
    """A plugin-type that provides plan/apply basics"""

    cli_label = "Planner"

    @tags(publish_to_cli=False)
    def goal(self, **kwargs):
        """ """
        return planning.Goal(
            owner=f"{self.__class__.__module__}.{self.__class__.__name__}", **kwargs
        )

    @property
    def Plan(self):
        return planning.Plan

    def plan(self, config=None) -> planning.Plan:
        """
        Creates a plan for this plugin
        """
        # app.manager.status_bar.update(app='PLAN')
        app.status_bar.update(
            app="Pynchon::PLAN", stage=f"plugin:{self.__class__.name}"
        )
        plan = self.Plan()
        app.status_bar.update(app="Pynchon", stage=f"{len(plan)}")
        return plan

    def apply(self, plan: planning.Plan = None) -> planning.ApplyResults:
        """
        Executes the plan for this plugin
        """
        cls_name = self.__class__.name
        msg = f"Applying for plugin '{cls_name}'"
        app.status_bar.update(
            app="Pynchon::APPLY", stage=f"plugin:{self.__class__.name}"
        )
        plan = plan or self.plan()
        goals = getattr(plan, "goals", plan)
        results = []
        total = len(goals)
        LOGGER.critical(f"{msg} ({total} goals)")
        git = self.siblings["git"]
        for i, action_item in enumerate(goals):
            app.status_bar.update(stage=f"{action_item}")
            cmd = action_item.command
            ordering = f"  {i+1}/{total}"
            prev_changes = git.modified
            invocation = invoke(cmd)
            rsrc_path = abcs.Path(action_item.resource).absolute()
            next_changes = [path.absolute() for path in git.modified]
            changed = all(
                [
                    rsrc_path in next_changes,
                    # rsrc_path not in prev_changes,
                ]
            )
            # raise Exception([changed,action_item.resource, next_changes])
            tmp = planning.Action(
                ok=invocation.succeeded,
                ordering=ordering,
                error="" if invocation.succeeded else invocation.stderr,
                # log=invocation.succeeded and invocation.stderr else None,
                owner=action_item.owner,
                command=action_item.command,
                resource=action_item.resource,
                type=action_item.type,
                changed=changed,
            )
            lme.CONSOLE.print(tmp)
            results.append(tmp)
        results = planning.ApplyResults(results)
        # write status event (used by the app-console)
        app.status_bar.update(
            app="Pynchon::HOOKS",
            stage=f"{cls_name}",
        )
        resources = list({r.resource for r in results})
        LOGGER.critical(f"Finished apply ({len(resources)} resources)")
        hooks = self.apply_hooks
        if hooks:
            self.logger.warning(
                f"{self.__class__} is dispatching {len(hooks)} hooks: {hooks}"
            )
            hook_results = []
            for hook in hooks:
                hook_results.append(self.run_hook(hook, results))
        else:
            self.logger.warning("No applicable hooks were found")
        return results

    def _validate_hooks(self, hooks):
        # FIXME: validation elsewhere
        for x in hooks:
            assert isinstance(x, (str,))
            assert " " not in x
            assert "_" not in x
            assert x.strip()

    @memoized_property
    def apply_hooks(self):
        """ """
        hooks = [x for x in self.hooks if x.split("-")[-1] == "apply"]
        apply_hooks = self["apply_hooks"::[]]
        hooks += [
            x + ("-apply" if not x.endswith("-apply") else "") for x in apply_hooks
        ]
        hooks = list(set(hooks))
        self._validate_hooks(hooks)
        return hooks

    @memoized_property
    def hooks(self):
        """ """
        hooks = self["hooks"::[]]
        self._validate_hooks(hooks)
        return hooks

    def _hook_open_after_apply(self, result: planning.ApplyResults) -> bool:
        """ """
        changes = list({r.resource for r in result})
        changes = [abcs.Path(rsrc) for rsrc in changes]
        changes = [rsrc for rsrc in changes if not rsrc.is_dir()]
        self.logger.warning(f"Opening {len(changes)} changed resources.")
        docs_plugin = self if self.name == "docs" else self.siblings["docs"]
        for ch in changes:
            docs_plugin.open(ch)
        return True

    @typing.validate_arguments
    def run_hook(self, hook_name: str, results: planning.ApplyResults):
        """
        :param hook_name: str:
        :param results: planning.ApplyResults:
        """

        class HookNotFound(Exception):
            pass

        class HookFailed(RuntimeError):
            pass

        norml_hook_name = hook_name.replace("-", "_")
        fxn_name = f"_hook_{norml_hook_name}"
        hook_fxn = getattr(self, fxn_name, None)
        if hook_fxn is None:
            err = [self.__class__, [hook_name, fxn_name]]
            self.logger.critical(err)
            raise HookNotFound(err)
        hook_result = hook_fxn(results)
        self.logger.critical(hook_result)
        return hook_result


class ShyPlanner(AbstractPlanner):
    """ShyPlanner uses plan/apply workflows, but they must be
    executed directly.  ProjectPlugin (or any other parent plugins)
    won't include this as a sub-plan.


    """

    contribute_plan_apply = False


@tags(cli_label="Manager")
class Manager(ShyPlanner):
    cli_label = "Manager"


class ResourceManager(Manager):
    @property
    def changes(self):
        """Set(git_changes).intersection(plugin_resources)"""
        git = self.siblings["git"]
        changes = git.modified
        these_changes = set(changes).intersection(set(self.list(changes=False)))
        return dict(modified=list(these_changes))

    @tagging.tags(click_aliases=["ls"])
    @cli.click.option(
        "--changes",
        "-m",
        "changes",
        is_flag=True,
        default=False,
        help="returns the git-modified subset",
    )
    def list(self, changes: bool = False):
        """Lists resources associated with this plugin"""
        if changes:
            return self.changes["modified"]
        from pynchon import abcs
        from pynchon.util import files

        try:
            include_patterns = self["include_patterns"]
            root = self["root"]
        except (KeyError,) as exc:
            self.logger.critical(
                f"{self.__class__} tried to use self.list(), but does not follow protocol"
            )
            self.logger.critical(
                "self['include_patterns'] and self['root'] must both be defined!"
            )
            raise
        root = abcs.Path(root)
        # proot = self.project_config['pynchon']['root']
        tmp = [p for p in include_patterns if abcs.Path(p).is_absolute()]
        tmp += [root / p for p in include_patterns if not abcs.Path(p).is_absolute()]
        # tmp += [proot / p for p in include_patterns if not abcs.Path(p).is_absolute()]
        return files.find_globs(tmp)


class Planner(ShyPlanner):
    """Planner uses plan/apply workflows, and contributes it's plans
    to ProjectPlugin (or any other parent plugins).
    """

    contribute_plan_apply = True



""" pynchon.util.text CLI
"""

from pynchon.cli import common

from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)
entry = common.entry_for(__name__)

# if child folders define __main__, consume entry-groups from those places
common.load_groups_from_children(root="pynchon.util.text", parent=entry)

if __name__ == "__main__":
    entry()



""" pynchon.util.text.normalize """

import re

from pynchon.util import typing


def snake_case(name: str) -> str:
    """

    :param name: str:
    :param name: str:

    """
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


DEFAULT_NORMALIZATION_RULES = {" ": "_", "/": "_", "-": "_"}


def normalize(
    txt: str = "",
    post: typing.List[typing.Callable] = [
        lambda _: _.lower(),
        lambda _: re.sub("_+", "_", _),
    ],
    rules: typing.List[typing.Callable] = DEFAULT_NORMALIZATION_RULES,
) -> str:
    """normalizes input text, with support for parametric rules/post-processing

    :param txt: str:  (Default value = "")
    :param post: typing.List[typing.Callable]:  (Default value = [lambda _: _.lower())
    :param lambda: _: re.sub('_+':
    :param txt: str:  (Default value = "")
    :param post: typing.List[typing.Callable]:  (Default value = [lambda _: _.lower())
    :param lambda _: re.sub('_+':
    :param '_':
    :param _):
    :param ]:
    :param rules: typing.List[typing.Callable]:  (Default value = DEFAULT_NORMALIZATION_RULES)

    """
    tmp = txt
    for k, v in rules.items():
        tmp = tmp.replace(k, v)
    for fxn in post:
        tmp = fxn(tmp)
    return tmp



""" pynchon.util.text.dumps """

import json as modjson

import yaml as modyaml
from fleks import models

from pynchon import abcs
from pynchon.util import lme, text

LOGGER = lme.get_logger(__name__)

json = models.to_json
JSONEncoder = models.JSONEncoder
JSONEncoder.register_encoder(type=abcs.Path, fxn=lambda v: str(v))


def yaml(file=None, content=None, obj=None):
    """
    Parse JSON input and returns (or writes) YAML
    """
    content = content or text.loadf.loadf(file=file, content=content)
    obj = obj or modjson.loads(content)
    content = modyaml.dump(obj)
    return content



""" pynchon.util.files.diff
"""

import difflib

import fleks

from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


def diff_report(diff, logger=LOGGER.debug):
    """

    :param diff: param logger:  (Default value = LOGGER.debug)
    :param logger:  (Default value = LOGGER.debug)

    """
    # FIXME: use rich.syntax
    import pygments
    import pygments.lexers
    import pygments.formatters

    tmp = pygments.highlight(
        diff,
        lexer=pygments.lexers.get_lexer_by_name("udiff"),
        formatter=pygments.formatters.get_formatter_by_name("terminal16m"),
    )
    logger(f"scaffold drift: \n\n{tmp}\n\n")


@fleks.cli.arguments.file1
@fleks.cli.arguments.file2
def diff_percent(file1: str = None, file2: str = None):
    """calculates file-delta, returning a percentage

    :param file1: str:  (Default value = None)
    :param file2: str:  (Default value = None)
    """
    with open(file1) as src:
        with open(file2) as dest:
            src_c = src.read()
            dest_c = dest.read()
    sm = difflib.SequenceMatcher(None, src_c, dest_c)
    return 100 * (1.0 - sm.ratio())


@fleks.cli.arguments.file1
@fleks.cli.arguments.file2
def strdiff(str1: str = None, str2: str = None, n=1):
    """calculates a file-delta, returning a unified diff

    :param str1: str:  (Default value = None)
    :param str2: str:  (Default value = None)
    """
    xdiff = difflib.unified_diff(
        str1.split("\n"),
        str2.split("\n"),
        fromfile="Current",
        tofile="Proposed",
        lineterm="",
        n=n,
    )
    return "\n".join(xdiff)


str_diff = strdiff


@fleks.cli.arguments.file1
@fleks.cli.arguments.file2
def diff(file1: str = None, file2: str = None):
    """calculates a file-delta, returning a unified diff

    :param file1: str:  (Default value = None)
    :param file2: str:  (Default value = None)
    """
    with open(file1) as src:
        with open(file2) as dest:
            str1 = src.readlines()
            str2 = dest.readlines()
    return strdiff(str1, str2)


file_diff = diff



""" pynchon.plugins.dot
"""

from fleks import cli

from pynchon.util.os import invoke

from pynchon import abcs, api, models  # noqa
from pynchon.util import files, lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


class Dot(models.Planner):
    """Finds / Renders (graphviz) dot files for this project"""

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "dot"
        exclude_patterns: typing.List[str] = typing.Field(default=[])

    name = "dot"

    def list(self) -> typing.List[str]:
        """ """
        # config = config or self.project_config
        search = [
            abcs.Path(self.project_root).joinpath("**/*.dot"),
        ]
        self.logger.debug(f"search pattern is {search}")
        result = files.find_globs(search)
        self.logger.debug(f"found {len(result)} files (pre-filter)")
        excludes = self["exclude_patterns" :: self[:"globals.exclude_patterns":]]
        self.logger.debug(f"filtering search with {len(excludes)} excludes")
        result = [p for p in result if not p.match_any_glob(excludes)]
        self.logger.debug(f"found {len(result)} files (post-filter)")
        if not result:
            err = "jinja-plugin is included in this config, but found no .j2 files!"
            self.logger.critical(err)
        return result

    @cli.options.output
    @cli.click.option("--img", default="nshine/dot")
    @cli.click.option("--output-mode")
    @cli.click.argument("file", nargs=1)
    def render(
        self,
        img: str = "??",
        file: str = "",
        output_mode: str = "png",
        output: str = "",
    ):
        cmd = f"cat {file} | docker run --rm --entrypoint dot -i {img} -T{output_mode} > {output}"
        result = invoke(cmd, strict=True)
        # assert result.succeeded
        return result.succeeded

    def plan(
        self,
        config=None,
    ) -> models.Plan:
        plan = super(self.__class__, self).plan(config=config)
        self.logger.debug("planning for rendering for .dot graph files..")
        cmd_t = "pynchon dot render {rsrc} --output {output} --output-mode png"
        for rsrc in self.list():
            plan.append(
                self.goal(
                    resource=rsrc,
                    command=cmd_t.format(
                        rsrc=rsrc, output=str(rsrc)[: str(rsrc).rfind(".")] + ".png"
                    ),
                    type="render",
                )
            )
        return plan



""" pynchon.core
"""

import os

from pynchon import abcs, constants
from pynchon.util import lme, typing

LOGGER = lme.get_logger(__name__)
DEFAULT_PLUGINS = constants.DEFAULT_PLUGINS

count = 0


def validate(kls=None, self=None, vdata=None):
    """ """
    global count
    count += 1
    if count > 1:
        raise Exception()

    def validate_plugins(plugin_list: typing.List = []):
        """ """
        defaults = set(constants.DEFAULT_PLUGINS)
        user_provided = set(plugin_list)
        intersection = defaults.intersection(user_provided)
        diff = defaults - intersection
        if diff:
            msg = "implied plugins are not mentioned explicitly!"
            vdata.warnings[msg].append(diff)

    def validate_config(k, v):
        if not isinstance(k, str) or (isinstance(k, str) and "{{" in k):
            raise ValueError(f"Top-level keys should be simple strings! {k}")
        if isinstance(v, str) and "{{" in v:
            raise ValueError(f"No templating in top level! {v}")
        raw_plugin_configs = {}
        if k == "plugins":
            validate_plugins(v)
        elif isinstance(v, (dict,)):
            raw_plugin_configs[k] = v
        else:
            # LOGGER.info(f'skipping validation for top-level `{k}` @ {v}')
            pass

    for k, v in dict(self).items():
        validate_config(k, v)


class Config(abcs.Config):
    """ """

    class Config:
        # https://github.com/pydantic/pydantic/discussions/5159
        # fields = {
        #     '_root': 'root',
        # }
        arbitrary_types_allowed = True
        frozen = True

    priority: typing.ClassVar[int] = 1
    config_key: typing.ClassVar[str] = "pynchon"

    __class_validators__ = []
    __instance_validators__ = [
        validate,
    ]

    # def __init__(self, **core_config):
    #     self.core_config = core_config
    #     if not core_config:
    #         LOGGER.critical("core config is empty!")
    #     super().__init__(**core_config)
    # def __init__(self, *args, **kwargs):
    #     # Do Pydantic validation
    #     super().__init__(*args, **kwargs)
    #     # Do things after Pydantic validation
    #     self.core_config=kwargs
    #     # if not any([self.alias, self.category, self.brand]):
    #     #     raise ValueError("No alias provided")

    @property
    def version(self) -> str:
        from pynchon import __version__

        return __version__

    @property
    def root(self) -> str:
        """{pynchon.root}:
        * user-config
        * os-env
        * {{git.root}}
        """
        from pynchon import config

        root = self.__dict__.get("root")
        root = root or os.environ.get("PYNCHON_ROOT")
        root = root or config.GIT.root
        root = root or self.working_dir
        return root and abcs.Path(root)

    # @tagging.tagged_property(conflict_strategy="override")
    @property
    def plugins(self):
        """{pynchon.plugins}:
        value here ultimately determines much of the
        rest of the apps bootstrap/runtime. value is
        always decided here, and must merge user-input
        from config files, plus any overrides on cli,
        plus pynchon's core set of default plugins.
        """
        from pynchon.config import MERGED_CONFIG_FILES

        plugins = MERGED_CONFIG_FILES.get("plugins", [])
        if not plugins:
            LOGGER.warning("no user-provided plugins found so far")
            # raise Exception(MERGED_CONFIG_FILES)
        return list(set(DEFAULT_PLUGINS + plugins))

    @property
    def working_dir(self):
        """working dir at the time of CLI invocation"""
        return abcs.Path(".").absolute()



""" pynchon.util.makefile CLI
"""

import shimport

from pynchon.cli import common

from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)
entry = common.entry_for(__name__)


(
    shimport.wrapper("pynchon.util.makefile")
    .prune(
        exclude_private=True,
        filter_module_origin=True,
        filter_instances=typing.FunctionType,
    )
    .map(lambda k, v: common.create_command(k, v, entry=entry))
)

if __name__ == "__main__":
    entry()



""" pynchon.codemod.commands.docstrings
"""

from .base import base  # noqa

# from .javadoc import *  # noqa

# import simple # noqa
# from libcst.codemod import CodemodContext, VisitorBasedCodemodCommand

# from .javadoc import javadoc

# class klass(base):
#     DESCRIPTION: str = """\tAdds or updates javadoc for Classes"""



""" pynchon.plugins.globals """

from pynchon import abcs, models

from pynchon.util import lme, typing  # noqa


class Globals(models.Provider):
    """Context for pynchon globals"""

    class config_class(abcs.Config):
        """ """

        config_key: typing.ClassVar[str] = "globals"
        exclude_patterns: typing.List[str] = typing.Field(default=[])

    priority = 2
    name = "globals"



""" pynchon.plugins.python.platform:
"""

import platform as stdlib_platform

from fleks import tagging

from pynchon import abcs, cli, models
from pynchon.util import lme, python, typing
from pynchon.util.os import invoke

LOGGER = lme.get_logger(__name__)


@tagging.tags(click_aliases=["py"])
class PythonPlatform(models.Planner):
    """
    Code transformation and docs-generation utilities for python projects.
    """

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "python"
        libcst: typing.Dict[str, typing.Any] = typing.Field(default={})

        @property
        def version(self):
            return stdlib_platform.python_version()

        @property
        def is_package(self) -> bool:
            return python.is_package(".")

        @property
        def package(self) -> typing.Dict:
            """ """
            if self.is_package:
                return PackageConfig()
            else:
                return {}

    cli_name = "python"

    priority = 2
    name = "python"

    @cli.click.group
    def bootstrap(self):
        """helpers for bootstrapping python projects"""

    @bootstrap.command("libcst")
    def bootstrap_libcst(self):
        """bootstrap .libcst.codemod.yaml"""
        # @cli.click.option('--libcst',help='bootstrap .libcst.codemod.yaml', default=False, is_flag=True)

    # @bootstrap.command("new")
    # @cli.click.argument("name")
    # def bootstrap_new(self):
    #     """shortcut for `pynchon cut new py/NAME`"""

    # def plan(self):
    #     plan = super(self.__class__, self).plan()
    #     libcst_config = self["libcst"]
    #     if libcst_config:
    #         plan.append(self._goal_libcst_refresh(libcst_config))
    #     return plan

    # @cli.click.group("src")
    # def src(self):
    #     """Generates code for python modules, packages, etc"""

    # @src.command
    @tagging.tags(click_aliases=["src.sorted"])
    def sorted(self):
        """Sorts code-ordering with `ssort`"""
        plan = super(self.__class__, self).plan()
        src_root = self[:"src.root":]
        plan.append(
            self.goal(
                type="code-gen",
                resource=src_root,
                command=f"pip install ssort==0.11.6 && ssort {src_root}",
            )
        )
        return self.apply(plan)


class PackageConfig(abcs.Config):
    """WARNING: `parent` below prevents moving this class elsewhere"""

    parent = PythonPlatform.config_class
    config_key: typing.ClassVar[str] = "package"

    @property
    def name(self) -> str:
        """ """
        from pynchon.util import python

        result = python.load_setupcfg().get("metadata", {}).get("name")
        return result

    @property
    def console_scripts(self) -> str:
        """ """
        from pynchon.util import python

        return python.load_entrypoints(python.load_setupcfg())

    @property
    def version(self) -> str:
        """ """
        if "version" not in self.__dict__:
            cmd = invoke("python setup.py --version 2>/dev/null", log_command=False)
            self.__dict__.update(version=cmd.succeeded and cmd.stdout.strip())
        return self.__dict__["version"]



""" pynchon.plugins.scaffolding
"""

from fleks import tagging

from pynchon import models
from pynchon.util import files, lme, typing
from pynchon.util.os import invoke

from .config import ScaffoldingConfig, ScaffoldingItem

tags = tagging.tags
LOGGER = lme.get_logger(__name__)


class Scaffolding(models.ShyPlanner):
    """Management tool for project boilerplate"""

    contribute_plan_apply = False
    priority = 3
    name = "scaffolding"
    cli_name = "scaffold"
    cli_label = "Manager"
    config_class = ScaffoldingConfig

    def match(self, pattern=None):
        """
        returns files that match for all scaffolds
        """
        if pattern:
            return files.find_globs([pattern], quiet=True)
        else:
            matches = []
            for scaffold in self.scaffolds:
                LOGGER.debug(
                    f"scaffold @ `{scaffold.pattern}` with scope {scaffold.scope}:"
                )
                matched_scaffold = ScaffoldingItem(
                    **{**scaffold, **dict(matches=self.match(pattern=scaffold.pattern))}
                )
                matches.append(matched_scaffold)
            return matches

    @property
    def matches(self):
        return self.match()

    @tags(click_aliases=["st", "status"])
    def stat(self):
        """status of current scaffolding"""
        ignore_keys = "diff".split()
        diff = self.diff(quiet=True)
        modified = [
            {k: v for k, v in s.items() if k not in ignore_keys}
            for s in diff["modified"]
        ]
        return dict(errors=diff["errors"], modified=modified)

    @property
    def scaffolds(self):
        """ """
        result = self.plugin_config.get("scaffolds", [])
        return [ScaffoldingItem(**x) for x in result]

    @tags(
        click_aliases=[
            "ls",
        ]
    )
    def list(self):
        """list available scaffolds"""
        return self.scaffolds

    def diff(self, quiet: bool = False):
        """diff with known scaffolding

        :param quiet: bool:  (Default value = False)
        :param quiet: bool:  (Default value = False)

        """
        result = dict(errors=[], modified=[])
        for matched_scaffold in self.matches:
            for fname in matched_scaffold.matches:
                if matched_scaffold.exists:
                    diff = invoke(f"diff {fname} {matched_scaffold.src}")
                    if diff.succeeded:
                        LOGGER.debug(f"no diff detected for {fname}")
                    else:
                        this_diff = files.diff(matched_scaffold.src, fname)
                        percent_diff = files.diff_percent(matched_scaffold.src, fname)
                        result["modified"].append(
                            dict(
                                src=matched_scaffold.src,
                                fname=fname,
                                percent_diff=f"{percent_diff}%",
                                # diff=this_diff,
                                # diff=diff.stdout
                            )
                        )
                        files.diff_report(this_diff, logger=LOGGER.debug)
                else:
                    result["errors"].append(fname)
        return result

    def plan(self, config=None) -> typing.List[str]:
        """

        :param config: Default value = None)

        """
        config or self.plugin_config
        plan = super().plan(config)
        for delta in self.diff()["modified"]:
            plan.append(
                models.Goal(
                    type="scaffold",
                    resource=delta["fname"],
                    command=f"cp {delta['src']} {delta['fname']}",
                )
            )
        return plan



""" pynchon.util.text.render

    Helpers for rendering content
"""

import os
import sys

from fleks.cli import click, options
from fleks.util.tagging import tags

from pynchon.util.os import invoke

from pynchon.util import lme, text, typing  # noqa

LOGGER = lme.get_logger(__name__)

from pynchon.api import render as api
from pynchon.util.text import loadf, loads


@typing.validate_arguments
def jinja(
    text: str = "",
    file: str = "?",
    context: dict = {},
    includes: typing.List[str] = [],
    strict: bool = True,
):
    """Renders jinja-templates (with support for includes)

    :param text: str:  (Default value = "")
    :param file: str:  (Default value = "?")
    :param context: dict:  (Default value = {})
    :param includes: typing.List[str]:  (Default value = [])
    :param strict: bool:  (Default value = True)
    :param text: str:  (Default value = "")
    :param file: str:  (Default value = "?")
    :param context: dict:  (Default value = {})
    :param includes: typing.List[str]:  (Default value = [])
    :param strict: bool:  (Default value = True)

    """
    import jinja2

    template = api.get_template_from_string(text, env=api.get_jinja_env(*includes))
    context = {
        # FIXME: try to santize this
        **dict(os.environ.items()),
        **dict(__template__=file),
        **context,
    }
    try:
        return template.render(**context)
    except (jinja2.exceptions.UndefinedError,) as exc:
        LOGGER.critical(f"Undefined exception: {exc}")
        LOGGER.critical(f"Jinja context: {list(context.keys())}")
        # import IPython; IPython.embed()
        raise
    except (jinja2.exceptions.TemplateNotFound,) as exc:
        LOGGER.critical(f"Template exception: {exc}")
        LOGGER.critical(f"Jinja-includes: {includes}")
        err = getattr(exc, "templates", exc.message)
        LOGGER.critical(f"Problem template: {err}")
        raise


@typing.validate_arguments
def jinja_loadf(
    file: str,
    context: typing.Dict = {},
    includes: typing.List[str] = [],
    strict: bool = True,
    quiet: bool = False,
) -> str:
    """

    :param file: str:
    :param context: typing.Dict:  (Default value = {})
    :param includes: typing.List[str]:  (Default value = [])
    :param strict: bool:  (Default value = True)
    :param quiet: bool:  (Default value = False)
    :param file: str:
    :param context: typing.Dict:  (Default value = {})
    :param includes: typing.List[str]:  (Default value = [])
    :param strict: bool:  (Default value = True)
    :param quiet: bool:  (Default value = False)

    """
    context = {} if context is None else context
    LOGGER.debug(f"Running with one file: {file} (strict={strict})")
    with open(file) as fhandle:
        content = fhandle.read()
    quiet and LOGGER.debug(f"render context: \n{text.to_json(context)}")
    tmp = list(context.keys())
    quiet and LOGGER.debug(f"Rendering with context:\n{text.to_json(tmp)}")
    content = jinja(text=content, file=file, context=context, includes=includes)
    # template = api.get_template(
    #     from_string=text,
    #     env=api.get_jinja_env(*includes))
    return content


@options.output
@options.should_print
@options.includes
# @options.context
# @options.context_file
@click.option("--context", help="context literal.  must be JSON")
@click.option("--context-file", help="context file.  must be JSON")
@click.argument("file", nargs=1)
@tags(
    click_aliases=["jinja"],
)
def jinja_file(
    file: str,
    output: typing.StringMaybe = "",
    should_print: typing.Bool = False,
    context: typing.Dict = {},
    context_file: typing.Dict = {},
    includes: typing.List[str] = [],
    strict: bool = True,
) -> str:
    """Renders jinja2 file (supports includes, custom filters)"""
    if isinstance(context, (str,)):
        LOGGER.warning("provided `context` is a string, loading it as JSON")
        context = loads.json(context)

    if context_file:
        assert not context
        context = loadf.json(context_file)

    content = jinja_loadf(
        file=file,
        context=context,
        includes=includes,
        strict=strict,
    )

    assert output
    output = os.path.splitext(output)
    if output[-1] == ".j2":
        output = output[0]
    else:
        output = "".join(output)
    LOGGER.warning(f"writing output to {output or sys.stdout.name}")
    from pynchon import abcs

    test = all([abcs.Path(output).exists(), output, output not in ["/dev/stdout", "-"]])
    if test:
        with open(output) as fhandle:
            before = fhandle.read()
    else:
        before = None
    fhandle = open(output, "w") if output else sys.stdout
    content = f"{content}\n"
    fhandle.write(content)
    fhandle.close()
    if before and content == before:
        LOGGER.critical(f"content in {output} did not change")
    return content


@options.output
@options.should_print
@click.option("--context", help="context file.  must be JSON")
@click.argument("file", nargs=1)
@tags(
    click_aliases=["j2"],
)
def j2cli(
    output: str, should_print: bool, file: str, context: str, format: str = "json"
) -> None:
    """A wrapper on the `j2` command (j2cli must be installed)
    Renders the named file, using the given context-file.

    NB: No support for jinja-includes or custom filters.

    :param output: str:
    :param should_print: bool:
    :param file: str:
    :param context: str:
    :param format: str:  (Default value = 'json')
    :param output: str:
    :param should_print: bool:
    :param file: str:
    :param context: str:
    :param format: str:  (Default value = 'json')

    """
    cmd = f"j2 --format {format} {file} {context}"
    result = invoke(cmd)
    if not result.succeeded:
        LOGGER.critical(f"failed to execute: {cmd}")
        raise SystemExit(1)
    result = result.stdout
    assert result
    tmp = file.replace(".j2", "")
    if tmp.endswith(".json") or tmp.endswith(".json5"):
        LOGGER.debug(f"target @ {file} appears to be specifying json.")
        LOGGER.debug("loading as if json5 before display..")
        result = text.loads.json5(content=result)
        result = text.to_json(result)
    msg = result
    print(msg, file=open(output, "w"))
    if should_print and output != "/dev/stdout":
        print(msg)


# assign utils back to sibling modules for convenience
# FIXME: is this smart?
loadf.jinja = jinja
loads.jinja = jinja



""" pynchon.plugins.pandoc
"""

from fleks import cli

from fleks.util import tagging  # noqa

from pynchon import abcs, events, models  # noqa
from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


class Pandoc(models.Planner):
    """Tool for working with Pandoc"""

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "pandoc"
        docker_image: str = typing.Field(default="pandoc/latex:latest")
        pdf_args: typing.List = typing.Field(
            default=["--toc", "--variable fontsize=10pt"]
        )

    name = "pandoc"
    cli_name = "pandoc"
    cli_label = "Tool"
    contribute_plan_apply = False

    # def plan(self, **kwargs):
    #     plan = super().plan()
    # @cli.click.flag("-b", "--bash", help="only bash codeblocks")
    @tagging.tags(click_aliases=["markdown.to-pdf"])
    @cli.options.output_file
    @cli.click.argument("file")
    def md_to_pdf(
        self,
        file: str = None,
        output: str = None,
    ):
        # -> ElementList:
        """
        Converts markdown files to PDF with pandoc
        """
        output = abcs.Path(output or f"{abcs.Path(file).stem}.pdf")
        pandoc_docker = self["docker_image"]
        docker_image = self["docker_image"]
        pdf_args = " ".join(self["pdf_args"])
        cmd = f"docker run -v `pwd`:/workspace -w /workspace {docker_image} {file} {pdf_args} -o {output}"
        plan = super().plan()
        plan.append(self.goal(resource=output.absolute(), type="render", command=cmd))
        return self.apply(plan=plan)



""" pynchon.util.docker
"""

import functools

from pynchon.util import lme, os

LOGGER = lme.get_logger(__name__)


class DockerCtx:
    def _run(
        self,
        command=None,
        entrypoint=None,
        working_dir=None,
        volumes={},
        workspace=False,
        **invoke_kwargs,
    ):
        entrypoint = entrypoint and f"--entrypoint {entrypoint}"
        entrypoint = entrypoint or ""
        if workspace:
            assert not working_dir
            working_dir = "/workspace"
            volumes["`pwd`"] = "/workspace"
        working_dir = working_dir and f"-w {working_dir}"
        working_dir = working_dir or ""
        volumes = [f"-v {k}:{v}" for k, v in volumes.items()]
        volumes = " ".join(volumes)
        volumes = volumes.strip()
        shcmd = f"docker run {working_dir} {volumes} {entrypoint} {self.tag} {command}"
        return os.invoke(shcmd, **invoke_kwargs)

    def __call__(
        self, command=None, entrypoint=None, script=None, **docker_or_invoke_kwargs
    ):
        """

        :param command: Default value = None)
        :param entrypoint: Default value = None)
        :param script: Default value = None)
        :param **docker_or_invoke_kwargs:

        """
        command = command or self.command
        script = script or self.script
        msg = self.init()
        LOGGER.debug("init result: {msg}")
        if script:
            fscript = ".tmp.docker.script"
            with open(fscript, "w") as fhandle:
                fhandle.write(script)
            entrypoint = entrypoint or (self.has_bash and "bash")
            entrypoint = entrypoint or (self.has_sh and "sh")
            if not entrypoint:
                raise Exception('container is missing "bash" & "sh"; cannot run script')
            return self._run(
                workspace=True,
                entrypoint=entrypoint,
                command=f"-x ./{fscript}",
                **docker_or_invoke_kwargs,
            )
        elif command:
            return self._run(command=command, workspace=True, **docker_or_invoke_kwargs)
        else:
            raise RuntimeError(f"not sure how to {self}.__call__ ")

    run = __call__

    def __init__(self, dockerfile: str = None):
        self.dockerfile = dockerfile and dockerfile.lstrip()
        self.command = None
        self.script = None

    @property
    def name(self):
        hash = ""  # FIXME
        raise NotImplementedError()
        # return f"same"

    @property
    def tag(self):
        return f"{self.name}:latest"

    @property
    def file(self):
        return f".tmp.dockerfile.{self.name}"

    @property
    def fhandle(self):
        return open(self.file, "w")

    def _run_bool(self, *args, **kwargs):
        return self._run(*args, **kwargs).succeeded

    def init(self):
        """ """
        if self.dockerfile:
            return self.init_dockerfile()
        elif self.tag:
            LOGGER.warning(
                f"i have tag {self.tag}, but can't be sure it's built or pulled"
            )
            return False
        else:
            raise RuntimeError(f"not sure how to init {self}")

    def init_dockerfile(self):
        """ """
        assert self.dockerfile
        LOGGER.warning(f"writing to {self.file}")
        print(self.dockerfile, file=self.fhandle)
        result = os.slurp_json(f"docker images {self.tag}" + " --format '{{json .}}'")
        if not result:
            LOGGER.warning(f'no container found for "{self.name}"')
            cmd = os.invoke(f"docker build --tag '{self.tag}' . -f {self.file}")
        else:
            LOGGER.warning("already have an image: ")
            LOGGER.warning(result)
        return result

    @property
    def has_sh(self):
        # return os.invoke(f'docker run --entrypoint sh {self.tag} -c "echo"').succeeded
        return self._run_bool(entrypoint="sh", command='-c "echo"')

    @property
    def has_bash(self):
        return self._run_bool(entrypoint="bash", command='-c "echo"')


def Dockerfile(txt, **kwargs):
    return functools.partial(DockerCtx, dockerfile=txt, **kwargs)()


if __name__ == "__main__":
    result = Dockerfile("""FROM python:3.8-alpine""").run(
        script='printf "{}"', load_json=True
    )
    result.json

    # def docker_fxn():
    #     """FROM python:3.8-alpine"""
    #     return "python --help"
    # Dockerfile(docker_fxn.__doc__)(docker_fxn)

    # class docker_cls:
    #     """FROM python:3.8-alpine"""
    #     script = "python --help"

    # os.docker_ctx(dockerfile='..', script='..')
    #
    # print(result)
    # print(result.stdout)
    # print(result.stderr)



""" pynchon.util.os.models
"""

# import os
# import subprocess
#
# from pynchon.util import lme
#
# LOGGER = lme.get_logger("pynchon.util.os")
#
# from fleks import app, meta
#
#
# class InvocationResult(meta.NamedTuple, metaclass=meta.namespace):
#     cmd: str = ""
#     stdin: str = ""
#     interactive: bool = False
#     large_output: bool = False
#     log_command: bool = True
#     environment: dict = {}
#     log_stdin: bool = True
#     system: bool = False
#     load_json: bool = False
#     json: dict = False
#     failed: bool = None
#     failure: bool = None
#     succeeded: bool = None
#     success: bool = None
#     stdout: str = ""
#     stderr: str = ""
#     pid: int = -1
#     shell: bool = False
#     strict: bool = False
#
#     def __rich__(self):
#         def status_string():
#             if self.succeeded is None:
#                 return "??"
#             return "[cyan]=> [green]ok" if self.succeeded else "[red]failed"
#
#         import shil
#
#         if self.log_command:
#             msg = f"running command: (system={self.system})\n  {self.cmd}"
#             fmt = shil.shfmt(self.cmd)
#             LOGGER.warning(msg)
#             syntax = app.Syntax(f"{fmt}", "bash", line_numbers=False, word_wrap=True)
#             panel = app.Panel(
#                 syntax,
#                 title=(
#                     f"{self.__class__.__name__} from "
#                     f"pid {self.pid} {status_string()}"
#                 ),
#                 title_align="left",
#                 subtitle=app.Text("✔", style="green")
#                 if self.success
#                 else app.Text("❌", style="red"),
#             )
#             lme.CONSOLE.print(panel)
#
#
# class Invocation(meta.NamedTuple, metaclass=meta.namespace):
#     cmd: str = ""
#     stdin: str = ""
#     strict: bool = False
#     shell: bool = False
#     interactive: bool = False
#     large_output: bool = False
#     log_command: bool = True
#     environment: dict = {}
#     log_stdin: bool = True
#     system: bool = False
#     load_json: bool = False
#
#     def __call__(self):
#         LOGGER.warning(self.cmd)
#         if self.system:
#             assert not self.stdin and not self.interactive
#             error = os.system(self.cmd)
#             return InvocationResult(
#                 **{
#                     **self._dict,
#                     **dict(
#                         failed=bool(error),
#                         failure=bool(error),
#                         success=not bool(error),
#                         succeeded=not bool(error),
#                         stdout="<os.system>",
#                         stdin="<os.system>",
#                     ),
#                 }
#             )
#         exec_kwargs = dict(
#             shell=True,
#             env={**{k: v for k, v in os.environ.items()}, **self.environment},
#         )
#         if self.stdin:
#             msg = "command will receive pipe:\n{}"
#             self.log_stdin and LOGGER.debug(msg.format(self.stdin))
#             exec_kwargs.update(
#                 stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE
#             )
#             LOGGER.critical([self.cmd, exec_kwargs])
#             exec_cmd = subprocess.Popen(self.cmd, **exec_kwargs)
#             exec_cmd.stdin.write(self.stdin.encode("utf-8"))
#             exec_cmd.wait()
#             exec_cmd.stdin.close()
#         else:
#             if not self.interactive:
#                 exec_kwargs.update(stdout=subprocess.PIPE, stderr=subprocess.PIPE)
#             exec_cmd = subprocess.Popen(self.cmd, **exec_kwargs)
#             exec_cmd.wait()
#         if exec_cmd.stdout:
#             exec_cmd.hstdout = exec_cmd.stdout
#             exec_cmd.stdout = (
#                 "<LargeOutput>"
#                 if self.large_output
#                 else exec_cmd.stdout.read().decode("utf-8")
#             )
#             exec_cmd.hstdout.close()
#         else:
#             exec_cmd.stdout = "<Interactive>"
#         if exec_cmd.stderr:
#             exec_cmd.hstderr = exec_cmd.stderr
#             exec_cmd.stderr = exec_cmd.stderr.read().decode("utf-8")
#             exec_cmd.hstderr.close()
#         exec_cmd.failed = exec_cmd.returncode > 0
#         exec_cmd.succeeded = not exec_cmd.failed
#         exec_cmd.success = exec_cmd.succeeded
#         exec_cmd.failure = exec_cmd.failed
#         loaded_json = None
#         if self.load_json:
#             if exec_cmd.failed:
#                 err = f"{self} did not succeed; cannot return JSON from failure"
#                 LOGGER.critical(err)
#                 LOGGER.critical(exec_cmd.stderr)
#                 raise RuntimeError(err)
#             import json
#
#             try:
#                 loaded_json = json.loads(exec_cmd.stdout)
#             except (json.decoder.JSONDecodeError,) as exc:
#                 loaded_json = dict(error=str(exc))
#         if self.strict and not exec_cmd.succeeded:
#             LOGGER.critical(f"Invocation failed and strict={self.strict}")
#             # raise InvocationError
#             LOGGER.critical(exec_cmd.stderr)
#             raise RuntimeError(exec_cmd.stderr)
#         return InvocationResult(
#             **{
#                 **self._dict,
#                 **dict(
#                     pid=exec_cmd.pid,
#                     failed=exec_cmd.failed,
#                     failure=exec_cmd.failure,
#                     success=exec_cmd.success,
#                     succeeded=exec_cmd.succeeded,
#                     stdout=exec_cmd.stdout,
#                     stderr=exec_cmd.stderr,
#                     json=loaded_json,
#                 ),
#             }
#         )



""" pynchon.models.plugins.cli
"""

import functools

import shimport
from fleks import classproperty, tagging

from pynchon import api, cli, events  # noqa
from pynchon.bin import entry  # noqa
from pynchon.util import lme, typing  # noqa

from .pynchon import PynchonPlugin  # noqa

LOGGER = lme.get_logger(__name__)
IPython = shimport.lazy("IPython")
config_mod = shimport.lazy("pynchon.config")


@tagging.tags(cli_label="<<Default>>")
class CliPlugin(PynchonPlugin):
    cli_label = "<<Default>>"
    _finalized_click_groups = dict()

    @classproperty
    def cli_path(kls):
        return f"{kls.click_entry.name} {kls.click_group.name}"

    @classproperty
    def click_entry(kls):
        return entry

    @classproperty
    def click_group(kls):
        cached = kls._finalized_click_groups.get(kls, None)
        grp_name = getattr(kls, "cli_name", kls.name)
        if cached is not None:
            return cached

        def plugin_main():
            pass

        plugin_main.__doc__ = (kls.__doc__ or "").lstrip()  # .split("\n")[0]
        groop = cli.common.groop(
            grp_name,
            parent=kls.click_entry,
        )
        plugin_main = groop(plugin_main)
        kls._finalized_click_groups[kls] = plugin_main

        tags = tagging.tags.get(kls) or {}
        gr_aliases = list(set(tags.get("click_aliases", [])))
        for alias in gr_aliases:
            g2 = cli.click.group_copy(plugin_main, name=alias, hidden=True)
            kls.click_entry.add_command(g2)

        return plugin_main

    @classproperty
    @functools.lru_cache(maxsize=None)
    def click_commands(kls) -> typing.List[str]:
        return [
            name
            for name in dir(kls)
            if name not in kls.__class_properties__
            and all(
                [
                    name not in "click_entry click_group".split(),
                    isinstance(getattr(kls, name), (cli.click.Command,)),
                    not isinstance(getattr(kls, name), (cli.click.Group,)),
                ]
            )
        ]

    @classproperty
    @functools.lru_cache(maxsize=None)
    def click_subgroups(kls) -> typing.List[str]:
        return [
            name
            for name in dir(kls)
            if name not in kls.__class_properties__
            and all(
                [
                    name not in "click_entry click_group".split(),
                    isinstance(getattr(kls, name), (cli.click.Group,)),
                ]
            )
        ]

    @PynchonPlugin.classmethod_dispatch(cli.click.Group)
    def click_acquire(
        kls, group: cli.click.Group, copy: bool = False, **update_kwargs
    ):  # noqa F811
        """
        :param group: cli.click.Group:
        """
        parent = kls.click_group
        LOGGER.info(
            f"{kls.__name__} acquires group@`{group.name}` to: parent@`{parent.name}`"
        )
        if copy:
            group = cli.click.group_copy(group, **update_kwargs)
        return cli.click.group_merge(group, parent)

    @PynchonPlugin.classmethod_dispatch(typing.FunctionType)
    def click_acquire(  # noqa: F811
        kls,
        fxn: typing.FunctionType,
        copy: bool = False,
        **update_kwargs,
    ):  # noqa F811
        """
        :param kls:
        :param fxn: typing.FunctionType:
        """
        msg = f"{kls.__name__} acquires naked fxn: {fxn.__name__}"
        assert fxn.__annotations__
        cmd_name = f"{fxn.__name__}".replace("_", "-")
        kls.click_group.add_command(cli.click.command(cmd_name)(fxn))

    @PynchonPlugin.classmethod_dispatch(cli.click.Command)
    def click_acquire(  # noqa: F811
        kls, cmd: cli.click.Command, copy: bool = False, **update_kwargs
    ):
        """ """
        parent = kls.click_group
        LOGGER.info(f"{kls.__name__} acquires {cmd.name} to: group@{parent.name}")
        if copy:
            cmd = cli.click.group_copy(cmd, **update_kwargs)
        parent.add_command(cmd)
        return parent

    @PynchonPlugin.classmethod_dispatch(typing.MethodType)
    def click_acquire(  # noqa: F811
        kls,
        cmd: typing.MethodType,
        copy: bool = False,  # irrelevant here
        **update_kwargs,
    ):
        """ """
        okls = kls
        fxn = cmd
        tags = tagging.tags[fxn]
        hidden = tags.get("click_hidden", False)
        click_aliases = tags.get("click_aliases", [])
        # click_parent_plugin = tags.get("click_parent_plugin", None)
        publish_to_cli = tags.get("publish_to_cli", True)

        from pynchon.util.text import dumps  # noqa

        if not publish_to_cli:
            return

        def wrapper(*args, fxn=fxn, **kwargs):
            LOGGER.warning(f"calling {fxn.__name__} from wrapper")
            result = fxn(*args, **kwargs)

            rproto = getattr(result, "__rich__", None)
            if rproto:
                # LOGGER.warning(f"rproto {result}")
                from pynchon.util.lme import CONSOLE

                CONSOLE.print(rproto())
            print(dumps.json(result))

        commands = [
            kls.click_create_cmd(
                fxn,
                wrapper=wrapper,
                **{
                    **update_kwargs,
                    **dict(hidden=hidden, alias=None),
                },
            )
        ]

        assert isinstance(
            click_aliases, (list, tuple)
        ), f"expected list or tuple for `click_aliases`, got {type(click_aliases)} in context {locals()}"
        for alias in click_aliases:
            if "." in alias:
                click_parent_plugin, name = alias.split(".")
                assert (
                    click_parent_plugin and name
                ), f"failed unpacking {alias} {locals()}"
                # raise Exception(click_parent_plugin)
                update_kwargs.update(
                    name=name,
                )
                update_kwargs.update(via=kls)
                kls = kls.siblings[click_parent_plugin]
                normalized_subcommand_name = fxn.__name__.replace("_", "-")
                update_kwargs.update(
                    help=f"(alias for `{okls.click_entry.name} {okls.cli_name} {normalized_subcommand_name}`)"
                )
            tmp = kls.click_create_cmd(
                fxn,
                wrapper=wrapper,
                **{
                    **update_kwargs,
                    **dict(hidden=hidden, alias=alias),
                },
            )
            commands.append(tmp)
        return commands

    @classmethod
    def init_cli(kls) -> cli.click.Group:
        """ """
        events.lifecycle.send(kls, plugin="initializing CLI")
        Core = shimport.lazy("pynchon.plugins.core").Core
        if kls != Core:
            # FIXME: this is needed, .. but why?
            config_mod.finalize()

        obj = kls.instance
        if obj is None:
            err = f"{kls.__name__}.`instance` is not ready?"
            LOGGER.warning(err)
            raise ValueError(err)

        cli_commands = []

        for group_name in kls.click_subgroups:
            grp = getattr(obj, group_name)
            cli_commands.append(grp)
            kls.click_acquire(grp)
            tags = tagging.tags.get(grp) or {}
            click_aliases = tags.get("click_aliases", [])
            if click_aliases:
                raise NotImplementedError()

        # create commands from commands
        # these command-callbacks are bound to non-methods,
        # i.e. currently expecting `self` as a first argument
        for cmd_name in kls.click_commands:
            cmd = getattr(obj, cmd_name)
            if isinstance(cmd.callback, (typing.FunctionType,)):
                cmd.callback = typing.bind_method(cmd.callback, obj)
            cli_commands.append(cmd)
            kls.click_acquire(cmd)

        # create commands from methods
        for method_name in kls.__methods__:
            fxn = obj and getattr(obj, method_name, None)
            if fxn is None:
                msg = f"    retrieved empty `{method_name}` from {obj}!"
                LOGGER.critical(msg)
                raise TypeError(msg)
            result = kls.click_acquire(fxn)
            if result is not None:
                cli_commands += result
        msg = [cmd.name for cmd in cli_commands]
        if len(msg) > 1:
            events.lifecycle.send(kls, plugin=f"created {len(msg)} commands")
        kls.init_cli_children()
        return kls.click_group

    @classmethod
    def init_cli_children(kls):
        """ """
        cli_subsumes = getattr(kls, "cli_subsumes", [])
        cli_subsumes and LOGGER.info(
            f"{kls.__name__} honoring `cli_subsumes`:\n\t{cli_subsumes}"
        )
        for fxn in cli_subsumes:
            raise NotImplementedError(fxn)
            # kls.click_acquire(fxn)

    @classmethod
    def click_create_cmd(
        kls,
        fxn: typing.Callable,
        via: typing.Any = None,
        wrapper=None,
        alias: str = None,
        **click_kwargs,
    ) -> cli.click.Command:
        """ """
        assert fxn
        assert wrapper
        via = f"the {via.__name__} plugin" if via else ""
        name = click_kwargs.pop("name", alias or fxn.__name__)
        name = name.replace("_", "-")
        alt = f"(alias for `{alias}`)" if alias else ""
        alt = alt or (f"(via {via})" if via else "")
        default = fxn.__doc__ or ""
        alt2 = default.lstrip().split("\n")
        alt2 = alt2[:2] + [" .. "] if len(alt2) > 2 else alt2[:2]
        alt2 = "\n".join(alt2)
        help = click_kwargs.pop(
            "help",
            (alt if alt else alt2),
        )
        help = help.lstrip()
        cmd = cli.common.kommand(
            name, help=help, alias=alias, parent=kls.click_group, **click_kwargs
        )(wrapper)
        options = getattr(fxn, "__click_params__", [])
        cmd.params += options
        return cmd

    @tagging.tags(
        click_aliases=["sh"],
        click_hidden=True,
    )
    @cli.click.option("--command", "-c", default="")
    def shell(self, command: str = "") -> None:
        """Drop to debugging shell"""
        before = locals()
        if command:
            self.logger.warning(f"executing command: {command} ")
            return eval(command)
        else:
            IPython.embed()
        after = {k: v for k, v in locals().items() if k not in before}
        LOGGER.warning(f"namespace changes: {after}")



""" pynchon.models.plugins.tool """

from fleks import tagging

from pynchon import api, cli, events  # noqa
from pynchon.util import lme, typing  # noqa

from .cli import CliPlugin  # noqa

LOGGER = lme.get_logger(__name__)


@tagging.tags(cli_label="Tool")
class ToolPlugin(CliPlugin):
    """
    Tool plugins may have their own config,
    but generally should not need project-config.
    """

    cli_label = "Tool"
    contribute_plan_apply = False
    __class_validators__ = [
        # validators.require_conf_key,
        # validators.warn_config_kls,
    ]



""" pynchon.annotate """

import os
import inspect
import importlib

from pynchon import abcs
from pynchon.util import complexity, lme, typing

LOGGER = lme.get_logger(__name__)


def klass(name, kls) -> typing.NoneType:
    """annotates a class

    :param name: param kls:
    :param kls:

    """
    LOGGER.debug(f"annotating class: {name}")
    mod = importlib.import_module(kls.parent.canonical_path)
    kls._handle = getattr(mod, name)

    properties = []
    for x in dir(kls._handle):
        if x.startswith("_"):
            continue
        try:
            prop = getattr(kls._handle, x)
        except (Exception,) as exc:
            LOGGER.critical(exc)
            continue
        is_property = type(prop).__name__ == "property"
        if not is_property:
            continue
        fxn = prop.fget
        fxn_doc = fxn.__doc__ or ""
        try:
            fxn_sig = inspect.signature(fxn)
        except (ValueError,) as exc:
            LOGGER.warning(f"Could not retrieve function signature for {fxn}!")
            fxn_sig = None
        try:
            fxn_code = fxn.__code__
        except AttributeError:  # C extensions..
            fxn_code = None
        start = fxn_code.co_firstlineno if fxn_code else kls.lineno
        properties.append(
            dict(
                name=x,
                doc=fxn_doc,
                signature=fxn_sig,
                start=str(start),
                end="",  # FIXME
                annotation=fxn_sig
                and str(fxn_sig.return_annotation)
                .replace("<class '", "")
                .replace("'>", ""),
                fixme="FIXME" in fxn_doc,
            )
        )

    bases = []
    for x in kls._handle.__bases__:
        tmp = "[{name}]({link})"
        qname = x.__module__.replace(".", "")
        bname = x.__name__
        if "builtin" in qname:
            tmp = tmp.format(
                name=f"`__builtin__.{bname}`",
                link=f"https://docs.python.org/3/library/functions.html#{bname}",
            )
        else:
            tmp = tmp.format(name=bname, link=f"#{qname}")
        bases.append(tmp)

    try:
        kls_code = inspect.getsource(kls._handle)
    except (Exception,) as exc:
        LOGGER.critical(exc)
        kls_code = "?"
    try:
        kls_fname = inspect.getfile(kls._handle)
    except (Exception,) as exc:
        LOGGER.critical(exc)
        kls_fname = "?"
    kls._metadata = dict(
        bases=bases,
        code=kls_code,
        start=str(kls.lineno),
        end=str(kls.endlineno or ""),
        properties=properties,
    )

    kls._metadata.update(mccabe=complexity.complexity(kls_code, kls_fname))


def module(name, module, working_dir=None) -> None:
    """annotates a module

    :param name: param module:
    :param working_dir: Default value = None)
    :param module:

    """
    LOGGER.debug(f"annotating module: {name}")
    working_dir = (working_dir or abcs.Path(".")).absolute()
    tmp = module.filepath.relative_to(working_dir)
    module._metadata = dict(base_url=str(tmp))


def should_skip(name: str):
    """
    :param name: str:
    """
    # from pynchon.config import pynchon as pynchon_config
    from pynchon.plugins.util import get_plugin_obj

    should_skip = get_plugin_obj("python-api")["skip_private_methods"]
    should_skip = should_skip and name.startswith("_")
    LOGGER.debug(
        f"annotation for `{name}` exits early; `pynchon.api.skip_private_methods` is set and this looks private"
    )
    return should_skip


def function(name, fxn) -> None:
    """
    annotates a function

    :param name: param fxn:
    :param fxn:

    """
    LOGGER.debug(f"annotating fxn: {name}")
    fxn._metadata = dict()
    if should_skip(name):
        return
    mod = importlib.import_module(fxn.parent.canonical_path)
    handle = getattr(mod, name)
    fxn._handle = getattr(mod, name)
    fxn_doc = fxn._handle.__doc__ or ""
    try:
        fxn_fname = inspect.getfile(fxn._handle)
    except (Exception,) as exc:
        fxn_fname = "?"
        fxn_code = None
    else:
        fxn_fname = os.path.relpath(fxn_fname)
        fxn_code = inspect.getsource(fxn._handle)
    try:
        fxn_sig = inspect.signature(fxn._handle)
    except (Exception,) as exc:
        fxn_sig = None
    fixme_lines = [
        i + fxn.lineno for i, l in enumerate(fxn_doc.split("\n")) if "FIXME" in l
    ]
    fxn._metadata = dict(
        name=name,
        doc=fxn_doc,
        signature=fxn_sig,
        # start=str(fxn._handle.__code__.co_firstlineno),
        annotation=str(fxn_sig and fxn_sig.return_annotation or "")
        .replace("<class '", "")
        .replace("'>", ""),
        fixme=(
            [
                dict(
                    glyph=" 🚩has FIXMEs ",
                    link=f"/{fxn_fname}#L{fixme_lines[0]}",
                    hover=f"on lines {fixme_lines}",
                )
            ]
            if fixme_lines
            else []
        ),
    )
    fxn_code and fxn._metadata.update(mccabe=complexity.complexity(fxn_code, fxn_fname))



""" pynchon.abcs
"""

from fleks.config import Config as FleksConfig  # noqa

from pynchon.util import typing  # noqa

from .attrdict import AttrDict  # noqa
from .path import Path  # noqa

ResourceType = typing.Union[str, Path]


class Config(FleksConfig):
    apply_hooks: typing.List[str] = typing.Field(default=[])



""" pynchon.models.plugins
"""

import typing

import fleks
from fleks import tagging

from pynchon import abcs, api, cli, events  # noqa
from pynchon.util import lme, typing  # noqa

from . import validators  # noqa
from .cli import CliPlugin  # noqa
from .provider import Provider  # noqa
from .pynchon import PynchonPlugin  # noqa
from .tool import ToolPlugin  # noqa

LOGGER = lme.get_logger(__name__)
classproperty = fleks.util.typing.classproperty


class BasePlugin(CliPlugin):
    """The default plugin-type most new plugins will use"""

    priority = 10

    @property
    def working_dir(self):
        """ """
        return abcs.Path(".").absolute()

    @property
    def exclude_patterns(self):
        """
        ensures that `exclude_patterns` for any plugin should honor the global-excludes
        """
        from pynchon.plugins import util as plugin_util

        globals = plugin_util.get_plugin("globals").get_current_config()
        global_ex = globals["exclude_patterns"]
        my_ex = self.get("exclude_patterns", [])
        return list(set(global_ex + my_ex + ["**/pynchon/templates/includes/**"]))


@tagging.tags(cli_label="NameSpace")
class NameSpace(CliPlugin):
    """`CliNamespace` collects functionality
    from elsewhere under a single namespace


    """

    cli_label = "NameSpace"
    contribute_plan_apply = False
    priority = 1



""" pynchon.plugins.cicd
"""

import webbrowser

from pynchon import abcs, models

from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


class CiCd(models.Provider):
    """Context for CI/CD"""

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "cicd"
        url_base: typing.Optional[str] = typing.Field(default=None)
        url_deploy: typing.Optional[str] = typing.Field(default=None)
        url_build: typing.Optional[str] = typing.Field(default=None)

        @property
        def type(self):
            """"""
            from pynchon.config import src

            src_root = abcs.Path(src["root"])
            default = "unknown"
            file_mapper = dict(
                jenkins="Jenkinsfile*",
                github_actions=".github/workflows/",
                travis=".travis.yml",
            )
            for typ, pat in file_mapper.items():
                if list(src_root.glob(pat)):
                    return typ
            return default

    name = "cicd"

    def open(self):
        """Opens CI/CD URL for this project"""
        url = self["url_build" :: self["url_base"]]
        if not url:
            LOGGER.critical("could not determine cicd url for this project config")
            return False
        else:
            return webbrowser.open(url)

    # @cli.click.option()
    def trigger(self):
        """Triggers CI/CD job for this project"""
        url = self["url_build"]
        raise NotImplementedError()



""" pynchon.plugins.mermaid
"""

import os

from fleks import cli, tagging

from pynchon.util.os import invoke

from pynchon import abcs, events, models  # noqa
from pynchon.util import files, lme, typing  # noqa

LOGGER = lme.get_logger(__name__)
IMG = "ghcr.io/mermaid-js/mermaid-cli/mermaid-cli"


class Mermaid(models.Planner):
    """Mermaid Plugin"""

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "mermaid"
        apply_hooks: typing.List[str] = typing.Field(default=["open-after"])

    name = "mermaid"
    cli_name = "mermaid"

    @tagging.tags(click_aliases=["ls"])
    def list(self):
        """
        Find mermaid diagrams under `{{project_root}}/**/*.mmd`
        """
        includes = "**/*.mmd"
        search = [
            abcs.Path(self.project_root).joinpath(includes),
        ]
        self.logger.debug(f"search pattern is {search}")
        result = files.find_globs(search)
        return result

    @cli.options.output
    @cli.click.option("--img", default="nshine/dot")
    # @cli.click.option("--output-mode")
    @cli.click.argument("file", nargs=1)
    def render(
        self,
        img: str = "??",
        file: str = "",
        # output_mode: str = "",
        output: str = "",
    ):
        """
        Renders mermaid diagram to image
        """
        assert output
        wd = self.working_dir
        file = abcs.Path(file).absolute().relative_to(wd)
        output = abcs.Path(output)
        output = output.absolute().relative_to(wd)
        uid = os.getuid()
        cmd = (
            f"docker run -v {wd}:/workspace "
            f"-w /workspace -u {uid} {IMG} -i {file} "
            f"-o {output} --backgroundColor efefef"
        )
        LOGGER.critical(cmd)
        result = invoke(cmd, strict=True)
        return True

    @property
    def output_root(self):
        return abcs.Path(self[:"git.root":]) / "img"

    def plan(
        self,
        config=None,
    ) -> models.Plan:
        """Run planning for this plugin"""
        plan = super(self.__class__, self).plan(config=config)
        self.logger.debug("planning for rendering for .mmd mermaid files..")
        output_mode = "png"
        for inp in self.list():
            rsrc = inp.parents[0] / inp.stem
            rsrc = f"{rsrc}.{output_mode}"
            plan.append(
                self.goal(
                    resource=rsrc,
                    command=(
                        f"pynchon {self.cli_name} " f"render {inp} --output {rsrc} "
                    ),
                    type="render",
                )
            )
        return plan

    # cli_label = "Tool"
    # def run(self):


# mmdc -i input.mmd -o output.png -t dark -b transparent
# https://github.com/mermaid-js/mermaid-cli



""" pynchon.plugins.tests
"""

from fleks import typing

from pynchon.util import lme

from pynchon import abcs, events, models  # noqa

LOGGER = lme.get_logger(__name__)


#
# class TestSuite(typing.BaseModel):
#     suite_name: str = typing.Field(default=None)
#     root: abcs.ResourceType = typing.Field(default=None, help="")
#     container: typing.Dict=typing.Field(default={}, help="")
#
# class DocTestSuite(TestSuite):
#     pass
#
#
class TestConfig(abcs.Config):
    # class Config:
    # arbitrary_types_allowed = True
    # include = 'suites'.split()
    config_key: typing.ClassVar[str] = "tests"


#     markdown: bool = typing.Field(
#         default=True,
#         # help='Configuration for testing markdown under {{docs/root}}',
#     )
#     html: bool = typing.Field(
#         default=False,
#         help='Configuration for testing HTML under {{docs/root}}',
#     )
#     suites: typing.List[TestSuite] = typing.Field(default=[])
#
#     @property
#     def suites(self):
#         tmp = self.__dict__.get("suites", [])
#         suite_names = [x.name for x in tmp]
#         if self.markdown and "markdown" not in suite_names:
#             LOGGER.critical("`markdown` is set but suite not found.  adding it..")
#             from pynchon.config import git
#
#             tmp += [TestSuite(name="markdown", root=git.root)]
#         return tmp
#         #     coverage={},
#         #     suite_patterns=[],
#         #     # suites={
#         #     #     "{{tests.root}}/units/": {
#         #     #           name:...
#         #     #           descr:...
#         #     #           runner:...
#         #     #      }
#         #     # }
#         #     root=None,
#


class Tests(models.Planner):
    """Management tool for project tests"""

    config_class = TestConfig
    name = "test"
    cli_name = "test"

    def _test_md(self, fname):
        """ """
        cfg = self.config
        LOGGER.warning(f"testing markdown: {fname}")

    # @cli.click.flag(
    #     "--markdown",
    # )
    # @cli.click.flag(
    #     "--html",
    # )
    # @cli.click.option(
    #     "--suffix",
    # )
    # @cli.options.plan
    # @cli.click.argument('file',default=None, required=False)
    def docs(
        self,
        markdown: bool = False,
        html: bool = False,
        suffix: str = None,
        should_plan: bool = False,
        file: str = None,
    ):
        """
        Run doc-tests for this project
        """
        files = [file] if file else self.list()
        plan = self.Plan()
        dct = {}
        for file in files:
            sfx = abcs.Path(file).suffix
            if sfx not in dct:
                dct[sfx] = []
            dct[sfx] += [file]
        if markdown:
            assert not any([file, suffix, html])
            suffix = ".md"
            files = dct[".md"]
        elif html:
            assert not any([file, suffix, markdown])
            suffix = ".html"

        if suffix:
            assert not file
            dct = {suffix: dct[suffix]}
        tmp = {}
        for suffix in dct.keys():
            hname = f"_test_{suffix[1:].replace('.', '_')}"
            suffix_handler = getattr(self, hname, None)
            if suffix_handler is None:
                LOGGER.critical(
                    f"missing doctest handler for {suffix}, {self.__class__.__name__}.{hname} is not present"
                )
            else:
                tmp[suffix_handler] = dct[suffix]
        dct = tmp

        for hdlr, flist in dct.items():
            for fname in flist:
                plan.append(
                    self.goal(
                        type="doctest",
                        resource=fname,
                        # command=f"pynchon parse markdown"
                        callable=hdlr,
                    )
                )

        if should_plan:
            return plan



""" pynchon.codemod.docstrings.simple """

import inspect
from textwrap import dedent

import libcst as cst
import shimport
from libcst._nodes.statement import (BaseSuite, ConcatenatedString, Expr,
                                     Sequence, SimpleStatementLine,
                                     SimpleString)

from pynchon.util import lme, typing

from .base import base

LOGGER = lme.get_logger(__name__)


def is_param_doc(txt: str):
    """
    :param txt: str:
    """
    return txt.lstrip().startswith(":param")


def get_obj(
    mod=None,
    dotpath=None,
):
    assert mod, "mod not set.. check .libcst.codemod.yaml src root"
    bits = dotpath.split(".")
    obits = [x for x in bits]
    obj = shimport.import_module(mod)
    while bits:
        try:
            obj = getattr(obj, bits.pop(0))
        except:
            LOGGER.critical(f"cannot import {obits}: got to {obj}; unfinished={bits} ")
            return None
    LOGGER.warning(f"imported {obj}")
    return obj


def write_docstring(
    mod=None,
    dotpath=None,
    obj=None,
    docstring=None,
    # indent=''
):
    """
    :param mod=None:
    :param dotpath=None:
    :param docstring=None:
    """
    if not isinstance(obj, (typing.FunctionType,)):
        LOGGER.critical(f"{obj} is not a Function!")
        return None
    # try:
    #     default_docstring = docs_from_typing(
    #         obj, style="numpy",
    # remove_linebreak=True
    #     )
    # except (AttributeError,) as exc:
    #     return

    doc_actual = inspect.getdoc(obj)

    default_docstring = str(inspect.signature(obj))
    default_docstring = default_docstring[1:-1].split(", ")
    if doc_actual is not None:
        LOGGER.warning(f"doc-string for {obj} already exists; extending it..")
        default_docstring = [
            f":param {x}:"
            for x in default_docstring
            if not any(
                [z.lstrip().startswith(f":param {x}:") for z in doc_actual.split("\n")]
            )
        ]
        default_docstring = doc_actual.split("\n") + default_docstring
    else:
        default_docstring = [f":param {x}:" for x in default_docstring]
    default_docstring = list(sorted(default_docstring))
    # :param arg1: description
    # :type arg1: type description
    # :return: return description
    # :rtype: the return type description
    src = inspect.getsource(obj).split("\n")
    index = [i for i, x in enumerate(src) if x.endswith(":")][0]
    ctx_indent = src[index + 1][: len(src[index + 1]) - len(dedent(src[index + 1]))]
    default_docstring = reversed(sorted(default_docstring))
    default_docstring = "\n".join(default_docstring)
    default_docstring = f'"""\n{default_docstring.strip()}\n"""'
    default_docstring = [(ctx_indent) + x for x in default_docstring.split("\n")]
    default_docstring = "\n".join(default_docstring)
    # LOGGER.critical(f"{dotpath}:: suggest:\n\n{default_docstring}")
    dstart = index + 1
    dpre = src[dstart].lstrip().rstrip()
    dpre = '"""' if dpre.startswith('"""') else ""
    dpre = dpre or ("'''" if dpre.startswith("'''") else "")
    for i, x in enumerate(src[dstart + 1 :]):
        x = x.lstrip().strip()
        if x.endswith(dpre):
            dend = i + dstart
            break
    else:
        LOGGER.critical(f"couldn't retrieve comment from: {src}")
        return None
    print(
        [
            dstart,
            dend,
            dpre,
        ]
    )
    src = src[:dstart] + [default_docstring] + src[dend:]
    src = "\n" + "\n".join([x for x in src])
    src = dedent(src)
    # LOGGER.critical(src)
    return src


def _get_docstring(body):
    """
    body
    """
    """ """
    if isinstance(body, Sequence):
        if body:
            expr = body[0]
        else:
            return expr, None
    else:
        expr = body
    while isinstance(expr, (BaseSuite, SimpleStatementLine)):
        if len(expr.body) == 0:
            return expr.body, None
        expr = expr.body[0]
    if not isinstance(expr, Expr):
        return expr, None
    val = expr.value
    if isinstance(val, (SimpleString, ConcatenatedString)):
        evaluated_value = val.evaluated_value
    else:
        return expr, None

    if evaluated_value is None:
        return expr, evaluated_value
    else:
        return inspect.cleandoc(evaluated_value)


class klass(base):
    DESCRIPTION: str = """\n\tWhere missing, adds docstrings to classes"""
    # from libcst import parse_statement
    # default_docstring = f'"""  """'
    # lctx = f"{original_node.__class__.__name__} @ '{full_module_name}'"
    # raise Exception([ltx, original_node])


class module(base):
    DESCRIPTION: str = """\n\tWhere missing, adds docstrings to modules"""

    def leave_Module(
        self,
        original_node: cst.Module,
        updated_node: cst.Module,
    ) -> cst.Module:
        full_module_name = self.context.full_module_name
        default_docstring = f'""" {full_module_name} """'
        lctx = f"{original_node.__class__.__name__} @ '{full_module_name}'"
        try:
            base = updated_node.children[0]
        except IndexError:
            # LOGGER.critical(f"{lctx}: empty-file!")
            return updated_node.with_changes(
                body=[cst.parse_statement(default_docstring)]
            )
        try:
            expr = base.children[0]
            sstring = expr.children[0]
        except IndexError:
            # FIXME: module starts with whitespace?
            # LOGGER.critical(f"{lctx}: starts with whitespace?")
            return original_node
        sstring = updated_node.children[0].children[0].children.pop(0)
        tmp = sstring.value
        if not tmp.strip():
            return updated_node.with_changes(
                body=[cst.parse_statement(default_docstring)] + list(original_node.body)
            )
        else:
            return original_node


class function(base):
    DESCRIPTION: str = """\n\tWhere missing, adds docstrings to functions"""

    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)
        self.this_class_name = None
        self.visiting_functions = []

    def leave_ClassDef(
        self,
        original_node: cst.ClassDef,
        updated_node: cst.ClassDef,
    ) -> cst.ClassDef:
        self.this_class_name = None
        return original_node

    def visit_ClassDef(
        self,
        original_node: cst.ClassDef,
        # updated_node: cst.ClassDef,
    ) -> cst.ClassDef:
        self.this_class_name = original_node.name.value
        return original_node

    def visit_FunctionDef(
        self,
        original_node: cst.FunctionDef,
        # updated_node: cst.FunctionDef,
    ) -> cst.FunctionDef:
        self.visiting_functions.append(original_node.name.value)
        return original_node

    def leave_FunctionDef(
        self,
        original_node: cst.FunctionDef,
        updated_node: cst.FunctionDef,
    ) -> cst.FunctionDef:
        import os

        from pynchon import abcs, api

        this_fname = self.visiting_functions.pop(-1)
        if self.visiting_functions:
            # nested func-def
            LOGGER.critical(
                f"ignoring nested function {[self.visiting_functions,this_fname]}"
            )
            return original_node
        mod = self.context.full_module_name
        src_root = api.project.get_config()["src"]["root"]
        tmp = self.context.full_module_name.replace(".", os.path.sep)
        tmp = abcs.Path(tmp).absolute()
        mod = str(tmp.relative_to(src_root)).replace(os.path.sep, ".")
        dotpath = f"{self.this_class_name+'.' if self.this_class_name else '' }{original_node.name.value}"
        ctx = f"{mod}.{dotpath}"
        obj = get_obj(
            mod=mod,
            dotpath=dotpath,
        )
        expr, docstring = _get_docstring(original_node)
        if obj is None:
            LOGGER.critical(f"could not retrieve object for {[mod,dotpath]}")
            import IPython

            IPython.embed()
        if docstring is not None and docstring.strip():
            LOGGER.critical("docstring ok")
            return original_node
        elif not (docstring and docstring.strip()):
            LOGGER.critical(f"{ctx}:: docstring is empty!")
            txt = write_docstring(
                mod=mod,
                dotpath=dotpath,
                obj=obj,
                docstring=docstring,
                # indent=self.context.module.default_indent
            )
            if txt is None:
                LOGGER.warning(f"error importing {[mod,dotpath]} ?")
                return original_node
            else:
                try:
                    updated_node = cst.parse_statement(txt)
                except (cst._exceptions.ParserSyntaxError,) as exc:
                    LOGGER.critical("ParserSyntaxError for update @\n\n{txt}")
                    return original_node
                else:
                    LOGGER.critical("updated node")
                    return updated_node

        elif docstring is None:
            LOGGER.info(f"{ctx}:: no docstring!")
            return original_node
            # if src:
            #     try:
            #         return cst.parse_statement(
            #             src,
            #             # cst.PartialParserConfig(
            #             #     default_indent=self.context.module.default_indent
            #             # ),
            #         )
            #     except (Exception,) as exc:
            #         LOGGER.critical(exc)

        return original_node



""" pynchon.events
"""

import blinker  # noqa
from blinker import signal  # noqa

from pynchon.util import lme

LOGGER = lme.get_logger(__name__)


class Signal(blinker.base.Signal):
    def _cleanup_receiver(self, *args, **kwargs):
        try:
            return super()._cleanup_receiver(*args, **kwargs)
        except (Exception,) as exc:
            pass
            # LOGGER.critical(f"ignoring exception")


lifecycle = Signal("lifecyle")
bootstrap = Signal("bootstrap")


# FIXME: use multi-dispatch over kwargs and define `lifecyle` repeatedly
def lifecycle_plugin(sender, plugin):
    """ """
    if plugin:
        tmp = getattr(sender, "__name__", getattr(sender, "name", str(sender)))
        tmp = f"{tmp}: PLUGIN: {plugin}"
        lifecycle.send(sender, msg=plugin)


def lifecycle_config(sender, config):
    """ """
    if config:
        tmp = getattr(sender, "__name__", getattr(sender, "name", str(sender)))
        tmp = f"{tmp}: CONFIG: {config}"
        lifecycle.send(sender, msg=config)


def lifecycle_applying(sender, applying=None, **kwargs):
    """ """
    if applying:
        tmp = getattr(sender, "__name__", getattr(sender, "name", str(sender)))
        tmp = f"{tmp}: APPLY: {applying}"
        lifecycle.send(lifecycle_applying, stage=tmp)


def lifecycle_stage(sender, stage=None, **unused):  # noqa
    """
    :param sender:
    :param stage:  (Default value = None)
    """
    if stage:
        tmp = getattr(sender, "__name__", str(sender))
        from pynchon.app import app

        app.status_bar.update(stage=stage)


def lifecycle_msg(sender, msg=None, **unused):  # noqa
    """
    :param sender: param msg:  (Default value = None)
    :param msg:  (Default value = None)
    """
    if msg:
        tmp = getattr(sender, "name", getattr(sender, "__name__", str(sender)))
        LOGGER.info(f"lifecycle :{tmp}: {msg}")


# ATTACHED=[]
def _lifecycle(sender, **signals):
    """ """
    from pynchon import events as THIS

    for k in signals:
        # if k in ATTACHED:
        #     continue
        # LOGGER.critical(f'attaching {k}')
        dispatch_name = f"lifecycle_{k}"
        dispatch = getattr(THIS, dispatch_name, None)
        loc = locals()
        assert dispatch, f"could not find {dispatch_name} in {loc}"
        dispatch(sender, **signals)
        # ATTACHED.append(k)


lifecycle.connect(_lifecycle)



""" pynchon.cli.common:
    Common options/arguments and base classes for CLI
"""

import json
import functools

import shimport
from fleks.cli import click

from pynchon.util import lme, text, typing  # noqa

LOGGER = lme.get_logger(__name__)


def load_groups_from_children(root=None, parent=None):
    """ """
    shimport.wrap(root).filter_folder(include_main=True).prune(
        name_is="entry",  # default
    ).map(
        lambda ch: [
            ch.name.replace(".__main__", "").split(".")[-1],
            ch.namespace["entry"],
        ]
    ).starmap(
        lambda name, fxn: [
            setattr(fxn, "name", name),
            click.group_merge(fxn, parent),
        ]
    )


def entry_for(
    name,
):
    """ """
    # from pynchon import shimport
    import importlib

    unf = importlib.import_module(name)
    mdoc = unf.__doc__

    def entry() -> typing.NoneType:
        pass

    entry.__doc__ = (mdoc or "").lstrip().split("\n")[0]

    class Groop(click.Group):
        pass

    entry = click.group(name.split(".")[-1], cls=Groop)(entry)
    return entry


class kommand:
    """ """

    is_group = False

    def __init__(
        self,
        name=None,
        parent=None,
        arguments=[],
        alias=None,
        options=[],
        transformers=[],
        handlers=[],
        formatters={},
        cls=None,
        help=None,
        **click_kwargs,
    ):
        self.name = name
        self.alias = alias
        self.parent = parent or click
        self.options = options
        self.arguments = arguments
        self.click_kwargs = click_kwargs
        self.formatters = {**formatters, **dict(json=self.format_json)}
        self.transformers = sorted(
            transformers,
            key=lambda t: t.priority,
        )
        self.handlers = sorted(
            handlers
            + [
                # output_handler(parent=self),
                # stdout_handler(parent=self),
            ],
            key=lambda h: h.priority,
        )
        cls and click_kwargs.update(cls=cls)
        help and click_kwargs.update(help=help.lstrip())
        if not self.is_group:
            self.cmd = self.parent.command(self.name, **click_kwargs)
        else:
            self.cmd = self.parent.group(self.name, **click_kwargs)
        self.cmd.alias = alias
        self.logger = lme.get_logger(f"cmd[{name}]")

    def format_json(self, result):
        """ """
        self.logger.debug("Formatter for: `json`")
        return json.dumps(result, indent=2)

    def wrapper(self, *args, **call_kwargs):
        assert self.fxn

        @functools.wraps(self.fxn)
        def newf(*args, **call_kwargs):
            self.logger.info(f"Wrapping invocation: {self.parent.name}.{self.name}")
            call_kwargs and self.logger.debug(f" with: {call_kwargs}")
            result = self.fxn(*args, **call_kwargs)
            if result is not None:
                result and LOGGER.info(f"json conversion for type: {type(result)}")
                import rich

                tmp = text.to_json(result)
                # raise Exception('bonk')
                rich.print(tmp)
            return result

        return newf

    def __call__(self, fxn: typing.Callable):
        self.fxn = fxn
        assert fxn, "function cannot be None!"

        f = self.cmd(self.wrapper())
        tmp = [x.strip() for x in (f.help or fxn.__doc__ or "").split("\n")]
        f.help = " ".join(tmp).lstrip()

        for opt in self.options:
            f = opt(f)
        for arg in self.arguments:
            f = arg(f)
        return f


def create_command(_name: str, fxn: typing.Callable, entry=None):
    """
    FIXME: move to fleks?
    WARNING: do not change signature.  this function is used
    with `shimport.wrapper().map(...)` so it must accept
    (key,value) style arguments.
    """
    from fleks.util.tagging import tags

    out = []
    aliases = tags[fxn].get("click_aliases", []) + [fxn.__name__]
    for alias in aliases:
        out.append(
            kommand(
                name=alias.replace("_", "-"),
                parent=entry,
                help=(
                    fxn.__doc__
                    if alias == fxn.__name__
                    else f"alias for `{fxn.__name__}`"
                ),
            )(fxn)
        )
    return out


class groop(kommand):
    """# class BetterGroup(click.Group):
    #     def format_usage(self,ctx,formatter):
    #         super(BetterGroup,self).format_usage(ctx,formatter)
    #     def format_epilog(self,ctx,formatter):
    #         return super(BetterGroup,self).format_epilog(ctx,formatter)
    #     def format_help(self, ctx, formatter):
    #         super(BetterGroup,self).format_help(ctx,formatter)
    #     def format_options(self,ctx,formatter):
    #         super(BetterGroup,self).format_options(ctx,formatter)
    #     def format_help_text(self,ctx,formatter):
    #         super(BetterGroup,self).format_help_text(ctx,formatter)
    """

    is_group = True



""" pynchon.plugins.render
"""

from pynchon import models

from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


class Renderers(models.NameSpace):
    """Collects `render` commands from other plugins"""

    name = cli_name = "render"
    config_class = None
    # cli_subsumes: typing.List[str] = [
    #     'pynchon.util.text.render.__main__',
    # ]


#
#
# @kommand(
#     name="dot",
#     parent=PARENT,
#     options=[
#         options.output,
#         click.option(
#             "--open",
#             "open_after",
#             is_flag=True,
#             default=False,
#             help=(f"if true, opens the created file using {DEFAULT_OPENER}"),
#         ),
#     ],
#     arguments=[files_arg],
# )
#
# @kommand(
#     name="any",
#     parent=PARENT,
#     formatters=dict(
#         # markdown=pynchon.T_TOC_CLI,
#     ),
#     options=[
#         # options.file,
#         options.format,
#         # options.stdout,
#         options.output,
#     ],
# )
# def render_any(format, file, stdout, output):
#     """
#     Render files with given renderer
#     """
#     raise NotImplementedError()
